#!/usr/bin/env python
"""
Tools for creating vector layers.
"""

import os
from typing import List

from osgeo import gdal
from osgeo import ogr
from osgeo import osr

import tqdm

import rsgislib

gdal.UseExceptions()


def polygonise_raster_to_vec_lyr(
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str,
    input_img: str,
    img_band: int = 1,
    mask_img: str = None,
    mask_band: int = 1,
    replace_file: bool = True,
    replace_lyr: bool = True,
    pxl_val_fieldname: str = "PXLVAL",
    use_8_conn: bool = False,
):
    """
    A utility to polygonise a raster to a OGR vector layer.

    :param out_vec_file: is a string specifying the output vector file path. If it
                         exists it will be deleted and overwritten.
    :param out_vec_lyr: is a string with the name of the vector layer.
    :param out_format: is a string with the driver
    :param input_img: is a string specifying the input image file to be polygonised
    :param img_band: is an int specifying the image band to be
                     polygonised. (default = 1)
    :param mask_img: is an optional string mask file specifying a no data
                     mask (default = None)
    :param mask_band: is an int specifying the image band to be used the
                      mask (default = 1)
    :param replace_file: is a boolean specifying whether the vector file should be
                         replaced (i.e., overwritten). Default=True.
    :param replace_lyr: is a boolean specifying whether the vector layer should be
                        replaced (i.e., overwritten). Default=True.
    :param pxl_val_fieldname: is a string to specify the name of the output column
                              representing the pixel value within the input image.
    :param use_8_conn: is a bool specifying whether 8 connectedness or 4 connectedness
                       should be used (4 is RSGISLib/GDAL default)

    """
    gdalImgDS = gdal.Open(input_img)
    imgBand = gdalImgDS.GetRasterBand(img_band)
    imgsrs = osr.SpatialReference()
    imgsrs.ImportFromWkt(gdalImgDS.GetProjectionRef())

    gdalImgMaskDS = None
    imgMaskBand = None
    if mask_img is not None:
        gdalImgMaskDS = gdal.Open(mask_img)
        imgMaskBand = gdalImgMaskDS.GetRasterBand(mask_band)

    if os.path.exists(out_vec_file) and (not replace_file):
        vecDS = gdal.OpenEx(out_vec_file, gdal.GA_Update)
    else:
        outdriver = ogr.GetDriverByName(out_format)
        if os.path.exists(out_vec_file):
            outdriver.DeleteDataSource(out_format)
        vecDS = outdriver.CreateDataSource(out_vec_file)

    if vecDS is None:
        raise rsgislib.RSGISPyException(
            "Could not open or create '{}'".format(out_vec_file)
        )

    lcl_options = []
    if replace_lyr:
        lcl_options = ["OVERWRITE=YES"]

    out_lyr_obj = vecDS.CreateLayer(out_vec_lyr, srs=imgsrs, options=lcl_options)
    if out_lyr_obj is None:
        raise rsgislib.RSGISPyException(
            "Could not create layer: {}".format(out_vec_lyr)
        )

    newField = ogr.FieldDefn(pxl_val_fieldname, ogr.OFTInteger)
    out_lyr_obj.CreateField(newField)
    dstFieldIdx = out_lyr_obj.GetLayerDefn().GetFieldIndex(pxl_val_fieldname)

    try:
        import tqdm

        pbar = tqdm.tqdm(total=100)
        callback = lambda *args, **kw: pbar.update()
    except:
        callback = gdal.TermProgress

    options = list()
    if use_8_conn:
        options.append("8CONNECTED=8")

    print("Polygonising...")
    gdal.Polygonize(
        imgBand, imgMaskBand, out_lyr_obj, dstFieldIdx, options, callback=callback
    )
    print("Completed")
    out_lyr_obj.SyncToDisk()
    vecDS = None
    gdalImgDS = None
    if mask_img is not None:
        gdalImgMaskDS = None


def vectorise_pxls_to_pts(
    input_img: str,
    img_band: int,
    img_msk_val: int,
    out_vec_file: str,
    out_vec_lyr: str = None,
    out_format: str = "GPKG",
    out_epsg_code: int = None,
    del_exist_vec: bool = False,
):
    """
    Function which creates a new output vector file for the pixels within the input
    image file with the value specified. Pixel locations will be the centroid of
    the the pixel

    :param input_img: the input image
    :param img_band: the band within the image to use
    :param img_msk_val: the image value selecting the pixels to be converted to points
    :param out_vec_file: Output vector file
    :param out_vec_lyr: output vector layer name.
    :param out_format: output file format (default GPKG).
    :param out_epsg_code: optionally provide an EPSG code for the output layer. If None
                          then taken from input image.
    :param del_exist_vec: remove output file if it exists.

    """
    from rios import applier
    import geopandas
    import rsgislib.imageutils
    import rsgislib.vectorutils

    try:
        progress_bar = rsgislib.TQDMProgressBar()
    except:
        from rios import cuiprogress

        progress_bar = cuiprogress.GDALProgressBar()

    if os.path.exists(out_vec_file):
        if del_exist_vec:
            rsgislib.vectorutils.delete_vector_file(out_vec_file)
        else:
            raise rsgislib.RSGISPyException(
                "The output vector file ({}) already exists, "
                "remove it and re-run.".format(out_vec_file)
            )

    if out_epsg_code is None:
        out_epsg_code = rsgislib.imageutils.get_epsg_proj_from_img(input_img)

    if out_epsg_code is None:
        raise rsgislib.RSGISPyException(
            "The output ESPG code is None - tried to read from input image and "
            "returned None. Suggest providing the EPSG code to the function."
        )

    pt_x_lst = list()
    pt_y_lst = list()

    infiles = applier.FilenameAssociations()
    infiles.image = input_img
    outfiles = applier.FilenameAssociations()
    otherargs = applier.OtherInputs()
    otherargs.img_band = img_band
    otherargs.img_msk_val = img_msk_val
    otherargs.pt_x_lst = pt_x_lst
    otherargs.pt_y_lst = pt_y_lst
    aControls = applier.ApplierControls()
    aControls.progress = progress_bar

    def _getXYPxlLocs(info, inputs, outputs, otherargs):
        """
        This is an internal rios function
        """
        x_block, y_block = info.getBlockCoordArrays()
        msk_data = inputs.image[0, ...].flatten()
        x_block = x_block.flatten()
        y_block = y_block.flatten()

        x_block = x_block[msk_data == otherargs.img_msk_val]
        y_block = y_block[msk_data == otherargs.img_msk_val]

        otherargs.pt_x_lst.extend(x_block.tolist())
        otherargs.pt_y_lst.extend(y_block.tolist())

    applier.apply(_getXYPxlLocs, infiles, outfiles, otherargs, controls=aControls)

    if len(pt_x_lst) > 0:

        gdf = geopandas.GeoDataFrame(
            geometry=geopandas.points_from_xy(pt_x_lst, pt_y_lst),
            crs="EPSG:{}".format(out_epsg_code),
        )

        if out_format == "GPKG":
            if out_vec_lyr is None:
                raise rsgislib.RSGISPyException(
                    "If output format is GPKG then an output layer is required."
                )
            gdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
        else:
            gdf.to_file(out_vec_file, driver=out_format)


def extract_image_footprint(
    input_img: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str = "GPKG",
    tmp_dir: str = "tmp",
    reproj_to: str = None,
    no_data_val=None,
):
    """
    A function to extract an image footprint as a vector.

    :param input_img: the input image file for which the footprint will be extracted.
    :param out_vec_file: output vector file path and name.
    :param out_vec_lyr: output vector layer name.
    :param tmp_dir: temp directory which will be used during processing. It will be
                    created and deleted once processing complete.
    :param reproj_to: optional if not None then an ogr2ogr command will be
                      run and the input here is what is to go into the ogr2ogr
                      command after -t_srs. E.g., -t_srs epsg:4326
    """
    gdal.UseExceptions()
    import rsgislib.tools.utils
    import rsgislib.tools.filetools
    import rsgislib.imageutils
    import rsgislib.vectorutils
    import rsgislib.vectorattrs

    if not rsgislib.tools.filetools.does_path_exists_or_creatable(out_vec_file):
        raise rsgislib.RSGISPyException(
            f"Output file path is not creatable: {out_vec_file}"
        )

    uid_str = rsgislib.tools.utils.uid_generator()

    created_tmp = False
    if not os.path.exists(tmp_dir):
        os.makedirs(tmp_dir)
        created_tmp = True

    in_img_base = os.path.splitext(os.path.basename(input_img))[0]

    valid_out_img = os.path.join(tmp_dir, in_img_base + "_" + uid_str + "_validimg.kea")
    if no_data_val is None:
        no_data_val = rsgislib.imageutils.get_img_no_data_value(input_img)

    if no_data_val is None:
        raise rsgislib.RSGISPyException("A no data value has not been specified.")
    rsgislib.imageutils.gen_valid_mask(
        input_img, valid_out_img, gdalformat="KEA", no_data_val=no_data_val
    )

    out_vec_tmp_file = out_vec_file
    if not (reproj_to is None):
        out_vec_tmp_file = os.path.join(
            tmp_dir, in_img_base + "_" + uid_str + "_initVecOut.gpkg"
        )

    polygonise_raster_to_vec_lyr(
        out_vec_tmp_file, out_vec_lyr, out_format, valid_out_img, 1, valid_out_img, 1
    )

    ds = gdal.OpenEx(out_vec_tmp_file, gdal.OF_READONLY)
    if ds is None:
        raise rsgislib.RSGISPyException("Could not open '" + out_vec_tmp_file + "'")

    lyr = ds.GetLayerByName(out_vec_lyr)
    if lyr is None:
        raise rsgislib.RSGISPyException("Could not find layer '" + out_vec_lyr + "'")
    num_feats = lyr.GetFeatureCount()
    lyr = None
    ds = None

    file_name = []
    for i in range(num_feats):
        file_name.append(os.path.basename(input_img))
    rsgislib.vectorattrs.write_vec_column(
        out_vec_tmp_file, out_vec_lyr, "FileName", ogr.OFTString, file_name
    )

    if not (reproj_to is None):
        if os.path.exists(out_vec_file):
            rsgislib.vectorutils.delete_vector_file(out_vec_file)

        cmd = [
            "ogr2ogr",
            "-f",
            out_format,
            "-t_srs",
            reproj_to,
            out_vec_file,
            out_vec_tmp_file,
        ]
        print(cmd)
        try:
            import subprocess

            subprocess.run(cmd, check=True)
        except OSError as e:
            raise rsgislib.RSGISPyException(
                "Could not re-projection vector file: {}".format(cmd)
            )

    if created_tmp:
        import shutil

        shutil.rmtree(tmp_dir)
    else:
        if not (reproj_to is None):
            driver = ogr.GetDriverByName("ESRI Shapefile")
            driver.DeleteDataSource(out_vec_tmp_file)


def create_poly_vec_for_lst_bboxs(
    csv_file,
    out_vec_file,
    out_vec_lyr,
    out_format,
    epsg_code,
    min_x_col=0,
    max_x_col=1,
    min_y_col=2,
    max_y_col=3,
    ignore_rows=0,
    del_exist_vec=False,
):
    """
    This function takes a CSV file of bounding boxes (1 per line) and creates a
    polygon vector layer.

    :param csv_file: input CSV file.
    :param out_vec_file: output vector file
    :param out_vec_file: output vector layer
    :param out_format: output vector file format (e.g., GPKG)
    :param epsg_code: EPSG code specifying the projection of the data
                     (4326 is WSG84 Lat/Long).
    :param min_x_col: The index (starting at 0) for the column within the CSV file
                    for the minimum X coordinate.
    :param max_x_col: The index (starting at 0) for the column within the CSV file
                    for the maximum X coordinate.
    :param min_y_col: The index (starting at 0) for the column within the CSV file
                    for the minimum Y coordinate.
    :param max_y_col: The index (starting at 0) for the column within the CSV file
                    for the maximum Y coordinate.
    :param ignore_rows: The number of rows to ignore from the start of the CSV
                        file (i.e., column headings)
    :param del_exist_vec: If the output file already exists delete it before proceeding.
    """
    gdal.UseExceptions()
    try:
        if os.path.exists(out_vec_file):
            if del_exist_vec:
                driver = ogr.GetDriverByName(out_format)
                driver.DeleteDataSource(out_vec_file)
            else:
                raise rsgislib.RSGISPyException("Output file already exists")
        # Create the output Driver
        outDriver = ogr.GetDriverByName(out_format)
        # create the spatial reference, WGS84
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(int(epsg_code))
        # Create the output Shapefile
        outDataSource = outDriver.CreateDataSource(out_vec_file)
        outLayer = outDataSource.CreateLayer(out_vec_lyr, srs, geom_type=ogr.wkbPolygon)
        # Get the output Layer's Feature Definition
        featureDefn = outLayer.GetLayerDefn()

        dataFile = open(csv_file, "r")
        rowCount = 0
        for line in dataFile:
            if rowCount >= ignore_rows:
                line = line.strip()
                if line != "":
                    comps = line.split(",")
                    # Get values from CSV file.
                    minX = float(comps[min_x_col])
                    maxX = float(comps[max_x_col])
                    minY = float(comps[min_y_col])
                    maxY = float(comps[max_y_col])
                    # Create Linear Ring
                    ring = ogr.Geometry(ogr.wkbLinearRing)
                    ring.AddPoint(minX, maxY)
                    ring.AddPoint(maxX, maxY)
                    ring.AddPoint(maxX, minY)
                    ring.AddPoint(minX, minY)
                    ring.AddPoint(minX, maxY)
                    # Create polygon.
                    poly = ogr.Geometry(ogr.wkbPolygon)
                    poly.AddGeometry(ring)
                    # Add to output shapefile.
                    outFeature = ogr.Feature(featureDefn)
                    outFeature.SetGeometry(poly)
                    outLayer.CreateFeature(outFeature)
                    outFeature = None
            rowCount = rowCount + 1
        dataFile.close()
        outDataSource = None
    except Exception as e:
        raise e


def define_grid(
    bbox,
    x_size,
    y_size,
    in_epsg_code,
    out_vec,
    out_vec_lyr,
    out_format="GPKG",
    out_epsg_code=None,
    utm_grid=False,
    utm_hemi=False,
):
    """
    Define a grid of bounding boxes for a specified bounding box. The output grid can
    be in a different projection to the inputted bounding box. Where a UTM grid is
    required and there are multiple UTM zones then the layer name will be appended
    with utmXX[n|s]. Note. this only works with formats such as GPKG which support
    multiple layers. A shapefile which only supports 1 layer will not work.

    :param bbox: a bounding box (xMin, xMax, yMin, yMax)
    :param x_size: Output grid size in X axis. If out_epsg_code or utm_grid defined
                   then the grid size needs to be in the output unit.
    :param y_size: Output grid size in Y axis. If out_epsg_code or utm_grid defined
                   then the grid size needs to be in the output unit.
    :param in_epsg_code: EPSG code for the projection of the bbox
    :param out_vec: output vector file.
    :param out_vec_lyr: output vector layer name.
    :param out_format: output vector file format (see OGR codes). Default is GPKG.
    :param out_epsg_code: if provided the output grid is reprojected to the projection
                          defined by this EPSG code. (note. the grid size needs to
                          the in the unit of this projection). Default is None.
    :param utm_grid: provide the output grid in UTM projection where grid might go
                     across multiple UTM zones. Default is False. grid size unit
                     should be metres.
    :param utm_hemi: if outputting a UTM projected grid then decided whether to use
                     hemispheres or otherwise. If False then everything will be
                     projected northern hemisphere (e.g., as with landsat or
                     sentinel-2). Default is False.
    """
    import rsgislib.tools.utm
    import rsgislib.tools.geometrytools

    if (out_epsg_code is not None) and utm_grid:
        raise rsgislib.RSGISPyException(
            "Cannot specify both new output projection and UTM grid."
        )
    elif utm_grid:
        wgs84_bbox = bbox
        if in_epsg_code != 4326:
            in_proj_obj = osr.SpatialReference()
            in_proj_obj.ImportFromEPSG(in_epsg_code)
            out_proj_obj = osr.SpatialReference()
            out_proj_obj.ImportFromEPSG(4326)
            wgs84_bbox = rsgislib.tools.geometrytools.reproj_bbox(
                bbox, in_proj_obj, out_proj_obj
            )

        multi_zones = False
        if (wgs84_bbox[0] < -180) and (wgs84_bbox[1] < -180):
            wgs84_bbox = [
                360 + wgs84_bbox[0],
                360 + wgs84_bbox[1],
                wgs84_bbox[2],
                wgs84_bbox[3],
            ]
        elif (wgs84_bbox[0] > 180) and (wgs84_bbox[1] > 180):
            wgs84_bbox = [
                360 - wgs84_bbox[0],
                360 - wgs84_bbox[1],
                wgs84_bbox[2],
                wgs84_bbox[3],
            ]
        elif (wgs84_bbox[0] < -180) or (wgs84_bbox[0] > 180):
            multi_zones = True

        if not multi_zones:
            utm_tl = rsgislib.tools.utm.from_latlon(wgs84_bbox[3], wgs84_bbox[0])
            utm_tr = rsgislib.tools.utm.from_latlon(wgs84_bbox[3], wgs84_bbox[1])
            utm_br = rsgislib.tools.utm.from_latlon(wgs84_bbox[2], wgs84_bbox[1])
            utm_bl = rsgislib.tools.utm.from_latlon(wgs84_bbox[2], wgs84_bbox[0])

            utm_top_hemi = "N"
            if utm_hemi and (wgs84_bbox[3] < 0):
                utm_top_hemi = "S"

        if (not multi_zones) and (utm_tl[2] == utm_tr[2] == utm_br[2] == utm_bl[2]):
            utm_zone = utm_tl[2]

            utm_proj_epsg = rsgislib.tools.utm.epsg_for_utm(utm_zone, utm_top_hemi)

            define_grid(
                bbox,
                x_size,
                y_size,
                in_epsg_code,
                out_vec,
                out_vec_lyr,
                out_format=out_format,
                out_epsg_code=utm_proj_epsg,
                utm_grid=False,
                utm_hemi=False,
            )
        else:
            multi_zones = True

            utm_zone_bboxs = []
            if wgs84_bbox[0] < -180:
                wgs84_bbox_W = [-180, wgs84_bbox[1], wgs84_bbox[2], wgs84_bbox[3]]
                wgs84_bbox_E = [360 + wgs84_bbox[1], 180, wgs84_bbox[2], wgs84_bbox[3]]

                utm_zone_bboxs = (
                    utm_zone_bboxs
                    + rsgislib.tools.utm.split_wgs84_bbox_utm_zones(wgs84_bbox_W)
                )
                utm_zone_bboxs = (
                    utm_zone_bboxs
                    + rsgislib.tools.utm.split_wgs84_bbox_utm_zones(wgs84_bbox_E)
                )

            elif wgs84_bbox[0] > 180:
                wgs84_bbox_W = [wgs84_bbox[0], 180, wgs84_bbox[2], wgs84_bbox[3]]
                wgs84_bbox_E = [-180, 360 - wgs84_bbox[1], wgs84_bbox[2], wgs84_bbox[3]]

                utm_zone_bboxs = (
                    utm_zone_bboxs
                    + rsgislib.tools.utm.split_wgs84_bbox_utm_zones(wgs84_bbox_W)
                )
                utm_zone_bboxs = (
                    utm_zone_bboxs
                    + rsgislib.tools.utm.split_wgs84_bbox_utm_zones(wgs84_bbox_E)
                )

            else:
                utm_zone_bboxs = rsgislib.tools.utm.split_wgs84_bbox_utm_zones(
                    wgs84_bbox
                )

            in_proj_obj = osr.SpatialReference()
            in_proj_obj.ImportFromEPSG(4326)

            first = True
            for zone_roi in utm_zone_bboxs:
                utm_top_hemi = "N"
                if utm_hemi:
                    if zone_roi[1][3] < 0:
                        utm_top_hemi = "S"

                utm_proj_epsg = int(
                    rsgislib.tools.utm.epsg_for_utm(zone_roi[0], utm_top_hemi)
                )

                out_proj_obj = osr.SpatialReference()
                out_proj_obj.ImportFromEPSG(utm_proj_epsg)

                utm_bbox = rsgislib.tools.geometrytools.reproj_bbox(
                    zone_roi[1], in_proj_obj, out_proj_obj
                )
                bboxs = rsgislib.tools.geometrytools.get_bbox_grid(
                    utm_bbox, x_size, y_size
                )

                utm_out_vec_lyr = out_vec_lyr + "_utm{0}{1}".format(
                    zone_roi[0], utm_top_hemi.lower()
                )
                create_poly_vec_bboxs(
                    out_vec,
                    utm_out_vec_lyr,
                    out_format,
                    utm_proj_epsg,
                    bboxs,
                    overwrite=first,
                )
                first = False
    else:
        if out_epsg_code is not None:
            in_proj_obj = osr.SpatialReference()
            in_proj_obj.ImportFromEPSG(in_epsg_code)
            out_proj_obj = osr.SpatialReference()
            out_proj_obj.ImportFromEPSG(out_epsg_code)
            proj_bbox = rsgislib.tools.geometrytools.reproj_bbox(
                bbox, in_proj_obj, out_proj_obj
            )
        else:
            proj_bbox = bbox

        bboxs = rsgislib.tools.geometrytools.get_bbox_grid(proj_bbox, x_size, y_size)

        if out_epsg_code is None:
            create_poly_vec_bboxs(out_vec, out_vec_lyr, out_format, in_epsg_code, bboxs)
        else:
            create_poly_vec_bboxs(
                out_vec, out_vec_lyr, out_format, out_epsg_code, bboxs
            )


def create_poly_vec_bboxs(
    vec_file,
    vec_lyr,
    out_format,
    epsg_code,
    bboxs,
    atts=None,
    att_types=None,
    overwrite=True,
):
    """
    This function creates a set of polygons for a set of bounding boxes.
    When creating an attribute the available data types are ogr.OFTString,
    ogr.OFTInteger, ogr.OFTReal

    :param vec_file: output vector file/path
    :param vec_lyr: output vector layer
    :param out_format: the output vector layer type.
    :param epsg_code: EPSG code specifying the projection of the data (e.g.,
                      4326 is WSG84 Lat/Long).
    :param bboxs: is a list of bounding boxes ([xMin, xMax, yMin, yMax]) to be
                  saved to the output vector.
    :param atts: is a dict of lists of attributes with the same length as the bboxs
                 list. The dict should be named the same as the attTypes['names'] list.
    :param att_types: is a dict with a list of attribute names (attTypes['names']) and
                     types (attTypes['types']). The list must be the same length as
                     one another and the number of atts. Example type: ogr.OFTString
    :param overwrite: overwrite the vector file specified if it exists. Use False
                      for GPKG where you want to add multiple layers.
    """
    try:
        gdal.UseExceptions()

        if os.path.exists(vec_file) and (not overwrite):
            # Open the output file.
            outDataSource = gdal.OpenEx(vec_file, gdal.GA_Update)
        else:
            # Create the output Driver
            outDriver = ogr.GetDriverByName(out_format)
            # Create the output vector file
            outDataSource = outDriver.CreateDataSource(vec_file)

        # create the spatial reference
        srs = osr.SpatialReference()
        srs.ImportFromEPSG(int(epsg_code))
        outLayer = outDataSource.CreateLayer(vec_lyr, srs, geom_type=ogr.wkbPolygon)

        addAtts = False
        if (atts is not None) and (att_types is not None):
            nAtts = 0
            if not "names" in att_types:
                raise rsgislib.RSGISPyException(
                    'attTypes must include a list for "names"'
                )
            nAtts = len(att_types["names"])
            if not "types" in att_types:
                raise rsgislib.RSGISPyException(
                    'attTypes must include a list for "types"'
                )
            if nAtts != len(att_types["types"]):
                raise rsgislib.RSGISPyException(
                    'attTypes "names" and "types" lists must be the same length.'
                )
            for i in range(nAtts):
                if att_types["names"][i] not in atts:
                    raise rsgislib.RSGISPyException(
                        '"{}" is not within atts'.format(att_types["names"][i])
                    )
                if len(atts[att_types["names"][i]]) != len(bboxs):
                    raise rsgislib.RSGISPyException(
                        '"{}" in atts does not have the same len as bboxs'.format(
                            att_types["names"][i]
                        )
                    )

            for i in range(nAtts):
                field_defn = ogr.FieldDefn(att_types["names"][i], att_types["types"][i])
                if outLayer.CreateField(field_defn) != 0:
                    raise rsgislib.RSGISPyException(
                        "Creating '" + att_types["names"][i] + "' field failed.\n"
                    )
            addAtts = True
        elif not ((atts is None) and (att_types is None)):
            raise rsgislib.RSGISPyException(
                "If atts or attTypes is not None then the other should also not be "
                "none and equalivent in length."
            )

        # Get the output Layer's Feature Definition
        featureDefn = outLayer.GetLayerDefn()

        openTransaction = False
        for n in range(len(bboxs)):
            if not openTransaction:
                outLayer.StartTransaction()
                openTransaction = True

            bbox = bboxs[n]
            # Create Linear Ring
            ring = ogr.Geometry(ogr.wkbLinearRing)
            ring.AddPoint(bbox[0], bbox[3])
            ring.AddPoint(bbox[1], bbox[3])
            ring.AddPoint(bbox[1], bbox[2])
            ring.AddPoint(bbox[0], bbox[2])
            ring.AddPoint(bbox[0], bbox[3])
            # Create polygon.
            poly = ogr.Geometry(ogr.wkbPolygon)
            poly.AddGeometry(ring)
            # Add to output shapefile.
            outFeature = ogr.Feature(featureDefn)
            outFeature.SetGeometry(poly)
            if addAtts:
                # Add Attributes
                for i in range(nAtts):
                    outFeature.SetField(
                        att_types["names"][i], atts[att_types["names"][i]][n]
                    )
            outLayer.CreateFeature(outFeature)
            outFeature = None
            if ((n % 20000) == 0) and openTransaction:
                outLayer.CommitTransaction()
                openTransaction = False

        if openTransaction:
            outLayer.CommitTransaction()
            openTransaction = False
        outDataSource = None
    except Exception as e:
        raise e


def write_pts_to_vec(
    out_vec_file,
    out_vec_lyr,
    out_format,
    epsg_code,
    pts_x,
    pts_y,
    atts=None,
    att_types=None,
    replace=True,
    file_opts=[],
    lyr_opts=[],
):
    """
    This function creates a set of polygons for a set of bounding boxes.
    When creating an attribute the available data types are ogr.OFTString,
    ogr.OFTInteger, ogr.OFTReal

    :param out_vec_file: output vector file/path
    :param out_vec_lyr: output vector layer
    :param out_format: the output vector layer type.
    :param epsg_code: EPSG code specifying the projection of the data
                      (e.g., 4326 is WSG84 Lat/Long).
    :param pts_x: is a list of x coordinates.
    :param pts_y: is a list of y coordinates.
    :param atts: is a dict of lists of attributes with the same length as the
                 ptsX & ptsY lists. The dict should be named the same as the
                 attTypes['names'] list.
    :param att_types: is a dict with a list of attribute names (attTypes['names'])
                      and types (attTypes['types']). The list must be the same length
                      as one another and the number of atts. Example type: ogr.OFTString
    :param replace: if the output vector file exists overwrite.
    :param file_opts: Options passed when creating the file. Default: [].
                      Common value might be ["OVERWRITE=YES"]
    :param lyr_opts: Options passed when create the layer Default: []. Common value
                     might be ["OVERWRITE=YES"]
    """
    import rsgislib.vectorutils
    from osgeo import ogr
    from osgeo import gdal
    from osgeo import osr

    try:
        if len(pts_x) != len(pts_y):
            raise rsgislib.RSGISPyException(
                "The X and Y coordinates lists are not the same length."
            )
        nPts = len(pts_x)

        gdal.UseExceptions()

        if os.path.exists(out_vec_file) and (not replace):
            vecDS = gdal.OpenEx(out_vec_file, gdal.GA_Update)
        else:
            if os.path.exists(out_vec_file):
                rsgislib.vectorutils.delete_vector_file(out_vec_file)
            outdriver = ogr.GetDriverByName(out_format)
            vecDS = outdriver.CreateDataSource(out_vec_file, options=file_opts)

        srs = osr.SpatialReference()
        srs.ImportFromEPSG(int(epsg_code))
        outLayer = vecDS.CreateLayer(
            out_vec_lyr, srs, geom_type=ogr.wkbPoint, options=lyr_opts
        )

        addAtts = False
        if (atts is not None) and (att_types is not None):
            nAtts = 0
            if not "names" in att_types:
                raise rsgislib.RSGISPyException(
                    'attTypes must include a list for "names"'
                )
            nAtts = len(att_types["names"])
            if not "types" in att_types:
                raise rsgislib.RSGISPyException(
                    'attTypes must include a list for "types"'
                )
            if nAtts != len(att_types["types"]):
                raise rsgislib.RSGISPyException(
                    'attTypes "names" and "types" lists must be the same length.'
                )
            for i in range(nAtts):
                if att_types["names"][i] not in atts:
                    raise rsgislib.RSGISPyException(
                        '"{}" is not within atts'.format(att_types["names"][i])
                    )
                if len(atts[att_types["names"][i]]) != len(pts_x):
                    raise rsgislib.RSGISPyException(
                        '"{}" in atts does not have the same len as pts_x'.format(
                            att_types["names"][i]
                        )
                    )

            for i in range(nAtts):
                field_defn = ogr.FieldDefn(att_types["names"][i], att_types["types"][i])
                if outLayer.CreateField(field_defn) != 0:
                    raise rsgislib.RSGISPyException(
                        "Creating '" + att_types["names"][i] + "' field failed.\n"
                    )
            addAtts = True
        elif not ((atts is None) and (att_types is None)):
            raise rsgislib.RSGISPyException(
                "If atts or attTypes is not None then the other should also "
                "not be none and equivlent in length."
            )

        # Get the output Layer's Feature Definition
        featureDefn = outLayer.GetLayerDefn()

        openTransaction = False
        for n in range(nPts):
            if not openTransaction:
                outLayer.StartTransaction()
                openTransaction = True
            # Create Point
            pt = ogr.Geometry(ogr.wkbPoint)
            pt.AddPoint(float(pts_x[n]), float(pts_y[n]))
            # Add to output shapefile.
            outFeature = ogr.Feature(featureDefn)
            outFeature.SetGeometry(pt)
            if addAtts:
                # Add Attributes
                for i in range(nAtts):
                    outFeature.SetField(
                        att_types["names"][i], atts[att_types["names"][i]][n]
                    )
            outLayer.CreateFeature(outFeature)
            outFeature = None
            if ((n % 20000) == 0) and openTransaction:
                outLayer.CommitTransaction()
                openTransaction = False

        if openTransaction:
            outLayer.CommitTransaction()
            openTransaction = False
        vecDS = None
    except Exception as e:
        raise e


def create_bboxs_for_pts(
    vec_file: str,
    vec_lyr: str,
    bbox_width: float,
    bbox_height: float,
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str = "GPKG",
    del_exist_vec: bool = False,
    epsg_code: int = None,
):
    """
    A function which takes a set of points (from the input vector layer) and creates
    a set of boxes with the same height and width, one for each point.

    Note, the geometry type for the input vector layer must be points.

    :param vec_file: the input vector file/path
    :param vec_lyr: the name of the input vector layer.
    :param bbox_width: width (in the units of the projection) for the output boxes
    :param bbox_height: height (in the units of the projection) for the output boxes
    :param out_vec_file: output vector file/path
    :param out_vec_lyr: output vector layer name
    :param out_format: output vector format.
    :param del_exist_vec: If the output file already exists delete it before proceeding.
    :param epsg_code: if not well defined specify the EPSG code for the projection.

    """
    import rsgislib.vectorutils

    if os.path.exists(out_vec_file):
        if del_exist_vec:
            rsgislib.vectorutils.delete_vector_file(out_vec_file)
        else:
            raise rsgislib.RSGISPyException(
                "The output vector file ({}) already exists, "
                "remove it and re-run.".format(out_vec_file)
            )

    h_width = bbox_width / 2.0
    h_height = bbox_height / 2.0

    vec_ds_obj = gdal.OpenEx(vec_file, gdal.OF_VECTOR)
    if vec_ds_obj is None:
        raise rsgislib.RSGISPyException("The input vector file could not be opened.")
    vec_lyr_obj = vec_ds_obj.GetLayer(vec_lyr)
    if vec_lyr_obj is None:
        raise rsgislib.RSGISPyException("The input vector layer could not be opened.")
    if epsg_code is None:
        epsg_code = rsgislib.vectorutils.get_proj_epsg_from_vec(vec_file, vec_lyr)

    n_feats = vec_lyr_obj.GetFeatureCount(True)
    pbar = tqdm.tqdm(total=n_feats)

    counter = 0
    in_feature = vec_lyr_obj.GetNextFeature()
    out_bboxs = list()
    while in_feature:

        geom = in_feature.GetGeometryRef()
        if geom is not None:
            pt_x = geom.GetX()
            pt_y = geom.GetY()

            x_min = pt_x - h_width
            x_max = pt_x + h_width
            y_min = pt_y - h_height
            y_max = pt_y + h_height

            out_bboxs.append([x_min, x_max, y_min, y_max])

        in_feature = vec_lyr_obj.GetNextFeature()
        counter = counter + 1
        pbar.update(1)

    pbar.close()
    vec_ds_obj = None

    create_poly_vec_bboxs(
        out_vec_file,
        out_vec_lyr,
        out_format,
        epsg_code,
        out_bboxs,
        atts=None,
        att_types=None,
        overwrite=False,
    )


def create_wgs84_vector_grid(out_vec_file:str, out_vec_lyr:str, out_format:str, grid_x:int, grid_y:int, bbox:List[float], overlap:float=None, tile_names_col:str='tile_names', tile_name_prefix:str=''):
    """
    A function which creates a regular grid across a defined area using the
    WGS84 (EPSG:4326) projection.

    :param out_vec_file: output vector file
    :param out_vec_lyr: output vector layer name
    :param out_format: the output vector file format.
    :param grid_x: the size in the x axis of the grid cells.
    :param grid_y: the size in the y axis of the grid cells.
    :param bbox: the area for which cells will be defined (MinX, MaxX, MinY, MaxY).
    :param overlap: the overlap added to each grid cell. If None then no overlap applied.
    :param tile_names_col: The output column name for the tile names.
    :param tile_name_prefix: A prefix for the tile names.

    """
    import math
    import rsgislib.tools.projection
    import rsgislib.vectorattrs

    epsg_code = 4326
    min_x = float(bbox[0])
    max_x = float(bbox[1])
    min_y = float(bbox[2])
    max_y = float(bbox[3])
    grid_x = float(grid_x)
    grid_y = float(grid_y)

    n_x_cells = math.floor((max_x - min_x) / grid_x)
    x_remain = (max_x - min_x) - (grid_x * n_x_cells)

    n_y_cells = math.floor((max_y - min_y) / grid_y)
    y_remain = (max_y - min_y) - (grid_y * n_y_cells)

    print("Cells: [{0}, {1}]".format(n_x_cells, n_y_cells))

    bboxs = []
    tile_names = []
    for i in range(n_y_cells):
        c_max_y = max_y - (i * grid_y)
        c_min_y = c_max_y - grid_y
        for j in range(n_x_cells):
            c_min_x = min_x + (j * grid_x)
            c_max_x = c_min_x + grid_x
            lat_lon_str_name = rsgislib.tools.projection.get_deg_coord_as_str(lat=c_min_x, lon=c_max_y, n_chars=4)
            tile_names.append(f"{tile_name_prefix}{lat_lon_str_name}")
            if overlap is None:
                bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])
            else:
                bboxs.append([c_min_x - overlap, c_max_x + overlap, c_min_y - overlap,
                              c_max_y + overlap])
        if x_remain > 0:
            c_min_x = min_x + (n_x_cells * grid_x)
            c_max_x = c_min_x + x_remain
            lat_lon_str_name = rsgislib.tools.projection.get_deg_coord_as_str(lat=c_min_x, lon=c_max_y, n_chars=4)
            tile_names.append(f"{tile_name_prefix}{lat_lon_str_name}")
            if overlap is None:
                bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])
            else:
                bboxs.append([c_min_x - overlap, c_max_x + overlap, c_min_y - overlap,
                              c_max_y + overlap])
    if y_remain > 0:
        c_max_y = max_y - (n_y_cells * grid_y)
        c_min_y = c_max_y - y_remain
        for j in range(n_x_cells):
            c_min_x = min_x + (j * grid_x)
            c_max_x = c_min_x + grid_x
            lat_lon_str_name = rsgislib.tools.projection.get_deg_coord_as_str(lat=c_min_x, lon=c_max_y, n_chars=4)
            tile_names.append(f"{tile_name_prefix}{lat_lon_str_name}")
            if overlap is None:
                bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])
            else:
                bboxs.append([c_min_x - overlap, c_max_x + overlap, c_min_y - overlap,
                              c_max_y + overlap])
        if x_remain > 0:
            c_min_x = min_x + (n_x_cells * grid_x)
            c_max_x = c_min_x + x_remain
            lat_lon_str_name = rsgislib.tools.projection.get_deg_coord_as_str(lat=c_min_x, lon=c_max_y, n_chars=4)
            tile_names.append(f"{tile_name_prefix}{lat_lon_str_name}")
            if overlap is None:
                bboxs.append([c_min_x, c_max_x, c_min_y, c_max_y])
            else:
                bboxs.append([c_min_x - overlap, c_max_x + overlap, c_min_y - overlap,
                              c_max_y + overlap])

    for bbox in bboxs:
        if bbox[2] < -180:
            bbox[2] = -180
        if bbox[3] > 180:
            bbox[3] = 180

    create_poly_vec_bboxs(out_vec_file, out_vec_lyr, out_format, epsg_code, bboxs)
    rsgislib.vectorattrs.write_vec_column(out_vec_file, out_vec_lyr, tile_names_col, ogr.OFTString, tile_names)





#!/usr/bin/env python
"""
The zonal stats module provides functions to perform pixel-in-polygon or
point-in-pixel analysis.

For undertaking a pixel-in-polygon analysis you need to consider the size of the
polygons with respect to the size of the pixels being intersected.

Where the pixels are small with respect to the polygons so there is at least one pixel
within the polygon then the best function to use is:

* rsgislib.zonalstats.calc_zonal_band_stats

If the pixels are large with respect to the polygons then use the following function
which intersects the polygon centroid.

* rsgislib.zonalstats.calc_zonal_poly_pts_band_stats

If the pixel size in between and/or polygons are varied in size such that it is not
certain that all polygons will contain a pixel then the following function will first
attempt to intersect the polygon with the pixels and if there is not a pixel within
the polygon then the centriod is used.

* rsgislib.zonalstats.calc_zonal_band_stats_test_poly_pts

"""
# import the C++ extension into this level
from ._zonalstats import *

import math
import sys
from typing import List

from osgeo import gdal
from osgeo import ogr
from osgeo import osr
import numpy
import tqdm

import rsgislib

gdal.UseExceptions()

"""
Alternatively, the other functions are slower to execute but have more options with
respect to the method of intersection. The options for intersection are:

* METHOD_POLYCONTAINSPIXEL = 0           # Polygon completely contains pixel
* METHOD_POLYCONTAINSPIXELCENTER = 1     # Pixel center is within the polygon
* METHOD_POLYOVERLAPSPIXEL = 2           # Polygon overlaps the pixel
* METHOD_POLYOVERLAPSORCONTAINSPIXEL = 3 # Polygon overlaps or contains the pixel
* METHOD_PIXELCONTAINSPOLY = 4           # Pixel contains the polygon
* METHOD_PIXELCONTAINSPOLYCENTER = 5     # Polygon center is within pixel
* METHOD_ADAPTIVE = 6                    # The method is chosen based on relative areas
                                         # of pixel and polygon.
* METHOD_ENVELOPE = 7                    # All pixels in polygon envelope chosen
* METHOD_PIXELAREAINPOLY = 8             # Percent of pixel area that is within
                                         # the polygon
* METHOD_POLYAREAINPIXEL = 9             # Percent of polygon area that is within pixel
"""

METHOD_POLYCONTAINSPIXEL = 0  # Polygon completely contains pixel
METHOD_POLYCONTAINSPIXELCENTER = 1  # Pixel center is within the polygon
METHOD_POLYOVERLAPSPIXEL = 2  # Polygon overlaps the pixel
METHOD_POLYOVERLAPSORCONTAINSPIXEL = 3  # Polygon overlaps or contains the pixel
METHOD_PIXELCONTAINSPOLY = 4  # Pixel contains the polygon
METHOD_PIXELCONTAINSPOLYCENTER = 5  # Polygon center is within pixel
METHOD_ADAPTIVE = 6  # The method is chosen based on relative areas
# of pixel and polygon.
METHOD_ENVELOPE = 7  # All pixels in polygon envelope chosen
METHOD_PIXELAREAINPOLY = 8  # Percent of pixel area that is within
# the polygon
METHOD_POLYAREAINPIXEL = 9  # Percent of polygon area that is within pixel


def calc_zonal_band_stats_file(
    vec_file,
    vec_lyr,
    input_img,
    img_band,
    min_thres,
    max_thres,
    out_no_data_val,
    min_field=None,
    max_field=None,
    mean_field=None,
    stddev_field=None,
    sum_field=None,
    count_field=None,
    mode_field=None,
    median_field=None,
    vec_def_epsg=None,
):
    """
    A function which calculates zonal statistics for a particular image band.
    If you know that the pixels in the values image are small with respect to
    the polygons then use this function.

    :param vec_file: input vector file
    :param vec_lyr: input vector layer within the input file which specifies the
                    features and where the output stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band
                     will be ignored.
    :param min_thres: a lower threshold for values which will be included in
                      the stats calculation.
    :param max_thres: a upper threshold for values which will be included in
                      the stats calculation.
    :param out_no_data_val: output no data value if no valid pixels are within
                            the polygon.
    :param min_field: the name of the field for the min value (None or not specified
                      to be ignored).
    :param max_field: the name of the field for the max value (None or not specified
                      to be ignored).
    :param mean_field: the name of the field for the mean value (None or not specified
                       to be ignored).
    :param stddev_field: the name of the field for the standard deviation value (None
                         or not specified to be ignored).
    :param sum_field: the name of the field for the sum value (None or not specified to
                      be ignored).
    :param count_field: the name of the field for the count (of number of pixels) value
                       (None or not specified to be ignored).
    :param mode_field: the name of the field for the mode value (None or not specified
                       to be ignored).
    :param median_field: the name of the field for the median value (None or not
                         specified to be ignored).
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is the
                         projection is not well defined within the inputted
                         vector layer.

    """
    try:
        vecDS = gdal.OpenEx(vec_file, gdal.OF_VECTOR | gdal.OF_UPDATE)
        if vecDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

        vec_lyr_obj = vecDS.GetLayerByName(vec_lyr)
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("Could not open layer '{}'".format(vec_lyr))

        calc_zonal_band_stats(
            vec_lyr_obj,
            input_img,
            img_band,
            min_thres,
            max_thres,
            out_no_data_val,
            min_field,
            max_field,
            mean_field,
            stddev_field,
            sum_field,
            count_field,
            mode_field,
            median_field,
            vec_def_epsg,
        )

        vecDS = None
    except Exception as e:
        print("Error Vector File: {}".format(vec_file), file=sys.stderr)
        print("Error Vector Layer: {}".format(vec_lyr), file=sys.stderr)
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def calc_zonal_band_stats(
    vec_lyr_obj,
    input_img,
    img_band,
    min_thres,
    max_thres,
    out_no_data_val,
    min_field=None,
    max_field=None,
    mean_field=None,
    stddev_field=None,
    sum_field=None,
    count_field=None,
    mode_field=None,
    median_field=None,
    vec_def_epsg=None,
):
    """
    A function which calculates zonal statistics for a particular image band.
    If you know that the pixels in the values image are small with respect to
    the polygons then use this function.

    :param vec_lyr_obj: OGR vector layer object containing the geometries being
                        processed and to which the stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band
                     will be ignored.
    :param min_thres: a lower threshold for values which will be included in
                      the stats calculation.
    :param max_thres: a upper threshold for values which will be included in
                      the stats calculation.
    :param out_no_data_val: output no data value if no valid pixels are within
                            the polygon.
    :param min_field: the name of the field for the min value (None or not specified
                      to be ignored).
    :param max_field: the name of the field for the max value (None or not specified
                      to be ignored).
    :param mean_field: the name of the field for the mean value (None or not specified
                       to be ignored).
    :param stddev_field: the name of the field for the standard deviation value (None
                         or not specified to be ignored).
    :param sum_field: the name of the field for the sum value (None or not specified to
                      be ignored).
    :param count_field: the name of the field for the count (of number of pixels) value
                       (None or not specified to be ignored).
    :param mode_field: the name of the field for the mode value (None or not specified
                       to be ignored).
    :param median_field: the name of the field for the median value (None or not
                         specified to be ignored).
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is the
                         projection is not well defined within the inputted
                         vector layer.

    """
    if mode_field is not None:
        import scipy.stats.mstats

    try:
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("The inputted vector layer was None")

        if (
            (min_field is None)
            and (max_field is None)
            and (mean_field is None)
            and (stddev_field is None)
            and (sum_field is None)
            and (count_field is None)
            and (mode_field is None)
            and (median_field is None)
        ):
            raise rsgislib.RSGISPyException(
                "At least one field needs to be specified for there is to an output."
            )

        imgDS = gdal.OpenEx(input_img, gdal.GA_ReadOnly)
        if imgDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(input_img))
        img_band_obj = imgDS.GetRasterBand(img_band)
        if img_band_obj is None:
            raise rsgislib.RSGISPyException(
                "Could not find image band '{}'".format(img_band)
            )
        imgGeoTrans = imgDS.GetGeoTransform()
        img_wkt_str = imgDS.GetProjection()
        img_spatial_ref = osr.SpatialReference()
        img_spatial_ref.ImportFromWkt(img_wkt_str)
        img_spatial_ref.AutoIdentifyEPSG()
        epsg_img_spatial = img_spatial_ref.GetAuthorityCode(None)

        pixel_width = imgGeoTrans[1]
        pixel_height = imgGeoTrans[5]

        imgSizeX = imgDS.RasterXSize
        imgSizeY = imgDS.RasterYSize

        imgNoDataVal = img_band_obj.GetNoDataValue()

        if vec_def_epsg is None:
            veclyr_spatial_ref = vec_lyr_obj.GetSpatialRef()
            if veclyr_spatial_ref is None:
                raise rsgislib.RSGISPyException(
                    "Could not retrieve a projection object from the vector layer - "
                    "projection might not be be defined."
                )
            epsg_vec_spatial = veclyr_spatial_ref.GetAuthorityCode(None)
        else:
            epsg_vec_spatial = vec_def_epsg
            veclyr_spatial_ref = osr.SpatialReference()
            veclyr_spatial_ref.ImportFromEPSG(int(vec_def_epsg))

        if epsg_vec_spatial != epsg_img_spatial:
            imgDS = None
            vecDS = None
            raise rsgislib.RSGISPyException(
                "Inputted raster and vector layers have different "
                "projections: ('{0}' '{1}') ".format("Vector Layer Provided", input_img)
            )

        veclyrDefn = vec_lyr_obj.GetLayerDefn()

        outFieldAtts = [
            min_field,
            max_field,
            mean_field,
            stddev_field,
            sum_field,
            count_field,
            mode_field,
            median_field,
        ]
        for outattname in outFieldAtts:
            if outattname is not None:
                found = False
                for i in range(veclyrDefn.GetFieldCount()):
                    if (
                        veclyrDefn.GetFieldDefn(i).GetName().lower()
                        == outattname.lower()
                    ):
                        found = True
                        break
                if not found:
                    vec_lyr_obj.CreateField(
                        ogr.FieldDefn(outattname.lower(), ogr.OFTReal)
                    )

        fieldAttIdxs = dict()
        for outattname in outFieldAtts:
            if outattname is not None:
                fieldAttIdxs[outattname] = vec_lyr_obj.FindFieldIndex(
                    outattname.lower(), True
                )

        vec_mem_drv = ogr.GetDriverByName("Memory")
        img_mem_drv = gdal.GetDriverByName("MEM")

        # Iterate through features.
        openTransaction = False
        transactionStep = 20000
        nextTransaction = transactionStep
        nFeats = vec_lyr_obj.GetFeatureCount(True)
        pbar = tqdm.tqdm(total=nFeats)
        counter = 0
        vec_lyr_obj.ResetReading()
        feat = vec_lyr_obj.GetNextFeature()
        while feat is not None:
            if not openTransaction:
                vec_lyr_obj.StartTransaction()
                openTransaction = True

            if feat is not None:
                feat_geom = feat.geometry()
                if feat_geom is not None:
                    feat_bbox = feat_geom.GetEnvelope()
                    havepxls = True

                    x1Sp = float(feat_bbox[0] - imgGeoTrans[0])
                    x2Sp = float(feat_bbox[1] - imgGeoTrans[0])
                    y1Sp = float(feat_bbox[3] - imgGeoTrans[3])
                    y2Sp = float(feat_bbox[2] - imgGeoTrans[3])

                    if x1Sp == 0.0:
                        x1 = 0
                    else:
                        x1 = int(x1Sp / pixel_width) - 1

                    if x2Sp == 0.0:
                        x2 = 0
                    else:
                        x2 = int(x2Sp / pixel_width) + 1

                    if y1Sp == 0.0:
                        y1 = 0
                    else:
                        y1 = int(y1Sp / pixel_height) - 1

                    if y2Sp == 0.0:
                        y2 = 0
                    else:
                        y2 = int(y2Sp / pixel_height) + 1

                    if x1 < 0:
                        x1 = 0
                    elif x1 >= imgSizeX:
                        x1 = imgSizeX - 1

                    if x2 < 0:
                        x2 = 0
                    elif x2 >= imgSizeX:
                        x2 = imgSizeX - 1

                    if y1 < 0:
                        y1 = 0
                    elif y1 >= imgSizeY:
                        y1 = imgSizeY - 1

                    if y2 < 0:
                        y2 = 0
                    elif y2 >= imgSizeY:
                        y2 = imgSizeY - 1

                    xsize = x2 - x1
                    ysize = y2 - y1

                    if (xsize == 0) or (ysize == 0):
                        havepxls = False

                    # Define the image ROI for the feature
                    src_offset = (x1, y1, xsize, ysize)

                    if havepxls:
                        # Read the band array.
                        src_array = img_band_obj.ReadAsArray(*src_offset)
                    else:
                        src_array = None

                    if (src_array is not None) and havepxls:

                        # calculate new geotransform of the feature subset
                        subGeoTrans = (
                            (imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1])),
                            imgGeoTrans[1],
                            0.0,
                            (imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5])),
                            0.0,
                            imgGeoTrans[5],
                        )

                        # Create a temporary vector layer in memory
                        vec_mem_ds = vec_mem_drv.CreateDataSource("out")
                        vec_mem_lyr = vec_mem_ds.CreateLayer(
                            "poly", veclyr_spatial_ref, ogr.wkbPolygon
                        )
                        vec_mem_lyr.CreateFeature(feat.Clone())

                        # Rasterize the feature.
                        img_tmp_ds = img_mem_drv.Create(
                            "", src_offset[2], src_offset[3], 1, gdal.GDT_Byte
                        )
                        img_tmp_ds.SetGeoTransform(subGeoTrans)
                        img_tmp_ds.SetProjection(img_wkt_str)
                        gdal.RasterizeLayer(
                            img_tmp_ds, [1], vec_mem_lyr, burn_values=[1]
                        )
                        rv_array = img_tmp_ds.ReadAsArray()

                        # Mask the data vals array to feature.
                        mask_arr = numpy.ones_like(src_array, dtype=numpy.uint8)
                        if imgNoDataVal is not None:
                            mask_arr[src_array == imgNoDataVal] = 0
                            mask_arr[rv_array == 0] = 0
                            mask_arr[src_array < min_thres] = 0
                            mask_arr[src_array > max_thres] = 0
                        else:
                            mask_arr[rv_array == 0] = 0
                            mask_arr[src_array < min_thres] = 0
                            mask_arr[src_array > max_thres] = 0
                        mask_arr = mask_arr.flatten()
                        src_array_flat = src_array.flatten()
                        src_array_flat = src_array_flat[mask_arr == 1]

                        if src_array_flat.shape[0] > 0:
                            if min_field is not None:
                                min_val = float(src_array_flat.min())
                                feat.SetField(fieldAttIdxs[min_field], min_val)
                            if max_field is not None:
                                max_val = float(src_array_flat.max())
                                feat.SetField(fieldAttIdxs[max_field], max_val)
                            if mean_field is not None:
                                mean_val = float(src_array_flat.mean())
                                feat.SetField(fieldAttIdxs[mean_field], mean_val)
                            if stddev_field is not None:
                                stddev_val = float(src_array_flat.std())
                                feat.SetField(fieldAttIdxs[stddev_field], stddev_val)
                            if sum_field is not None:
                                sum_val = float(src_array_flat.sum())
                                feat.SetField(fieldAttIdxs[sum_field], sum_val)
                            if count_field is not None:
                                count_val = float(src_array_flat.shape[0])
                                feat.SetField(fieldAttIdxs[count_field], count_val)
                            if mode_field is not None:
                                mode_val, mode_count = scipy.stats.mstats.mode(
                                    src_array_flat
                                )
                                mode_val = float(mode_val)
                                feat.SetField(fieldAttIdxs[mode_field], mode_val)
                            if median_field is not None:
                                median_val = float(numpy.ma.median(src_array_flat))
                                feat.SetField(fieldAttIdxs[median_field], median_val)
                        else:
                            if min_field is not None:
                                feat.SetField(fieldAttIdxs[min_field], out_no_data_val)
                            if max_field is not None:
                                feat.SetField(fieldAttIdxs[max_field], out_no_data_val)
                            if mean_field is not None:
                                feat.SetField(fieldAttIdxs[mean_field], out_no_data_val)
                            if stddev_field is not None:
                                feat.SetField(
                                    fieldAttIdxs[stddev_field], out_no_data_val
                                )
                            if sum_field is not None:
                                feat.SetField(fieldAttIdxs[sum_field], out_no_data_val)
                            if count_field is not None:
                                feat.SetField(
                                    fieldAttIdxs[count_field], out_no_data_val
                                )
                            if mode_field is not None:
                                feat.SetField(fieldAttIdxs[mode_field], out_no_data_val)
                            if median_field is not None:
                                feat.SetField(
                                    fieldAttIdxs[median_field], out_no_data_val
                                )
                        # Write the updated feature to the vector layer.
                        vec_lyr_obj.SetFeature(feat)

                        vec_mem_ds = None
                        img_tmp_ds = None

            if (counter == nextTransaction) and openTransaction:
                vec_lyr_obj.CommitTransaction()
                openTransaction = False
                nextTransaction = nextTransaction + transactionStep

            feat = vec_lyr_obj.GetNextFeature()
            counter = counter + 1
            pbar.update(counter)
        if openTransaction:
            vec_lyr_obj.CommitTransaction()
            openTransaction = False
        pbar.close()

        imgDS = None
    except Exception as e:
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def calc_zonal_poly_pts_band_stats_file(
    vec_file, vec_lyr, input_img, img_band, out_field, vec_def_epsg=None
):
    """
    A funtion which extracts zonal stats for a polygon using the polygon centroid.
    This is useful when you are intersecting a low resolution image with respect to
    the polygon resolution.

    :param vec_file: input vector file
    :param vec_lyr: input vector layer within the input file which specifies the
                    features and where the output stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band will
                     be ignored.
    :param out_field: output field name within the vector layer.
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is the
                         projection is not well defined within the inputted
                         vector layer.

    """
    try:
        vecDS = gdal.OpenEx(vec_file, gdal.OF_VECTOR | gdal.OF_UPDATE)
        if vecDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

        vec_lyr_obj = vecDS.GetLayerByName(vec_lyr)
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("Could not open layer '{}'".format(vec_lyr))

        calc_zonal_poly_pts_band_stats(
            vec_lyr_obj, input_img, img_band, out_field, vec_def_epsg
        )

        vecDS = None
    except Exception as e:
        print("Error Vector File: {}".format(vec_file), file=sys.stderr)
        print("Error Vector Layer: {}".format(vec_lyr), file=sys.stderr)
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def calc_zonal_poly_pts_band_stats(
    vec_lyr_obj, input_img, img_band, out_field, vec_def_epsg=None
):
    """
    A funtion which extracts zonal stats for a polygon using the polygon centroid.
    This is useful when you are intesecting a low resolution image with respect to
    the polygon resolution.

    :param vec_lyr_obj: OGR vector layer object containing the geometries being
                        processed and to which the stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band
                     will be ignored.
    :param out_field: output field name within the vector layer.
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is
                         the projection is not well defined within the inputted
                         vector layer.

    """
    try:
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("The inputted vector layer was None")

        imgDS = gdal.OpenEx(input_img, gdal.GA_ReadOnly)
        if imgDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(input_img))
        img_band_obj = imgDS.GetRasterBand(img_band)
        if img_band_obj is None:
            raise rsgislib.RSGISPyException(
                "Could not find image band '{}'".format(img_band)
            )
        imgGeoTrans = imgDS.GetGeoTransform()
        img_wkt_str = imgDS.GetProjection()
        img_spatial_ref = osr.SpatialReference()
        img_spatial_ref.ImportFromWkt(img_wkt_str)
        epsg_img_spatial = img_spatial_ref.GetAuthorityCode(None)

        pixel_width = imgGeoTrans[1]
        pixel_height = imgGeoTrans[5]

        imgSizeX = imgDS.RasterXSize
        imgSizeY = imgDS.RasterYSize

        if vec_def_epsg is None:
            veclyr_spatial_ref = vec_lyr_obj.GetSpatialRef()
            if veclyr_spatial_ref is None:
                raise rsgislib.RSGISPyException(
                    "Could not retrieve a projection object from the vector layer - "
                    "projection not might be be defined."
                )
            epsg_vec_spatial = veclyr_spatial_ref.GetAuthorityCode(None)
        else:
            epsg_vec_spatial = vec_def_epsg

        if epsg_vec_spatial != epsg_img_spatial:
            imgDS = None
            vecDS = None
            raise rsgislib.RSGISPyException(
                "Inputted raster and vector layers have different "
                "projections: ('{0}' '{1}') ".format("Vector Layer Provided", input_img)
            )

        veclyrDefn = vec_lyr_obj.GetLayerDefn()

        found = False
        for i in range(veclyrDefn.GetFieldCount()):
            if veclyrDefn.GetFieldDefn(i).GetName().lower() == out_field.lower():
                found = True
                break
        if not found:
            vec_lyr_obj.CreateField(ogr.FieldDefn(out_field.lower(), ogr.OFTReal))

        outfieldidx = vec_lyr_obj.FindFieldIndex(out_field.lower(), True)

        vec_mem_drv = ogr.GetDriverByName("Memory")
        img_mem_drv = gdal.GetDriverByName("MEM")

        # Iterate through features.
        openTransaction = False
        transactionStep = 20000
        nextTransaction = transactionStep
        nFeats = vec_lyr_obj.GetFeatureCount(True)
        pbar = tqdm.tqdm(total=nFeats)
        counter = 0
        vec_lyr_obj.ResetReading()
        feat = vec_lyr_obj.GetNextFeature()
        while feat is not None:
            if not openTransaction:
                vec_lyr_obj.StartTransaction()
                openTransaction = True

            if feat is not None:
                feat_geom = feat.geometry()
                if feat_geom is not None:
                    feat_bbox = feat_geom.GetEnvelope()
                    havepxls = True

                    x1Sp = float(feat_bbox[0] - imgGeoTrans[0])
                    x2Sp = float(feat_bbox[1] - imgGeoTrans[0])
                    y1Sp = float(feat_bbox[3] - imgGeoTrans[3])
                    y2Sp = float(feat_bbox[2] - imgGeoTrans[3])

                    if x1Sp == 0.0:
                        x1 = 0
                    else:
                        x1 = int(x1Sp / pixel_width) - 1

                    if x2Sp == 0.0:
                        x2 = 0
                    else:
                        x2 = int(x2Sp / pixel_width) + 1

                    if y1Sp == 0.0:
                        y1 = 0
                    else:
                        y1 = int(y1Sp / pixel_height) - 1

                    if y2Sp == 0.0:
                        y2 = 0
                    else:
                        y2 = int(y2Sp / pixel_height) + 1

                    if x1 < 0:
                        x1 = 0
                    elif x1 >= imgSizeX:
                        x1 = imgSizeX - 1

                    if x2 < 0:
                        x2 = 0
                    elif x2 >= imgSizeX:
                        x2 = imgSizeX - 1

                    if y1 < 0:
                        y1 = 0
                    elif y1 >= imgSizeY:
                        y1 = imgSizeY - 1

                    if y2 < 0:
                        y2 = 0
                    elif y2 >= imgSizeY:
                        y2 = imgSizeY - 1

                    xsize = x2 - x1
                    ysize = y2 - y1

                    if (xsize == 0) or (ysize == 0):
                        havepxls = False

                    # Define the image ROI for the feature
                    src_offset = (x1, y1, xsize, ysize)

                    if havepxls:
                        # Read the band array.
                        src_array = img_band_obj.ReadAsArray(*src_offset)
                    else:
                        src_array = None

                    if (src_array is not None) and havepxls:
                        subTLX = imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1])
                        subTLY = imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5])
                        resX = imgGeoTrans[1]
                        resY = imgGeoTrans[5]

                        ptx, pty, ptz = feat.GetGeometryRef().Centroid().GetPoint()

                        xOff = math.floor((ptx - subTLX) / resX)
                        yOff = math.floor((pty - subTLY) / resY)

                        if xOff < 0:
                            xOff = 0
                        if xOff >= xsize:
                            xOff = xsize - 1

                        if yOff < 0:
                            yOff = 0
                        if yOff >= ysize:
                            yOff = ysize - 1

                        out_val = float(src_array[yOff, xOff])
                        feat.SetField(outfieldidx, out_val)

                        vec_lyr_obj.SetFeature(feat)

                        vec_mem_ds = None
                        img_tmp_ds = None

            if (counter == nextTransaction) and openTransaction:
                vec_lyr_obj.CommitTransaction()
                openTransaction = False
                nextTransaction = nextTransaction + transactionStep

            feat = vec_lyr_obj.GetNextFeature()
            counter = counter + 1
            pbar.update(counter)

        if openTransaction:
            vec_lyr_obj.CommitTransaction()
            openTransaction = False
        pbar.close()

        imgDS = None
    except Exception as e:
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def calc_zonal_band_stats_test_poly_pts_file(
    vec_file,
    vec_lyr,
    input_img,
    img_band,
    min_thres,
    max_thres,
    out_no_data_val,
    percentile=None,
    percentile_field=None,
    min_field=None,
    max_field=None,
    mean_field=None,
    stddev_field=None,
    sum_field=None,
    count_field=None,
    mode_field=None,
    median_field=None,
    vec_def_epsg=None,
):
    """
    A function which calculates zonal statistics for a particular image band. If
    unsure then use this function. This function tests whether 1 or more pixels has
    been found within the polygon and if not then the centroid use used to find a
    value for the polygon.

    If you are unsure as to whether the pixels are small enough to be contained
    within all the polygons then use this function.

    :param vec_file: input vector file
    :param vec_lyr: input vector layer within the input file which specifies the
                    features and where the output stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band
                     will be ignored.
    :param min_thres: a lower threshold for values which will be included in
                      the stats calculation.
    :param max_thres: a upper threshold for values which will be included in
                      the stats calculation.
    :param out_no_data_val: output no data value if no valid pixels are within
                            the polygon.
    :param percentile: the percentile value to calculate (value between 0 and
                       100 inclusive).
    :param percentile_field: the name of the field for the percentile value (None or
                             not specified to be ignored).
    :param min_field: the name of the field for the min value (None or not specified
                      to be ignored).
    :param max_field: the name of the field for the max value (None or not specified
                      to be ignored).
    :param mean_field: the name of the field for the mean value (None or not specified
                       to be ignored).
    :param stddev_field: the name of the field for the standard deviation value (None
                         or not specified to be ignored).
    :param sum_field: the name of the field for the sum value (None or not specified to
                      be ignored).
    :param count_field: the name of the field for the count (of number of pixels) value
                       (None or not specified to be ignored).
    :param mode_field: the name of the field for the mode value (None or not specified
                       to be ignored).
    :param median_field: the name of the field for the median value (None or not
                         specified to be ignored).
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is the
                         projection is not well defined within the inputted
                         vector layer.

    """
    try:
        vecDS = gdal.OpenEx(vec_file, gdal.OF_VECTOR | gdal.OF_UPDATE)
        if vecDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

        vec_lyr_obj = vecDS.GetLayerByName(vec_lyr)
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("Could not open layer '{}'".format(vec_lyr))

        calc_zonal_band_stats_test_poly_pts(
            vec_lyr_obj,
            input_img,
            img_band,
            min_thres,
            max_thres,
            out_no_data_val,
            percentile,
            percentile_field,
            min_field,
            max_field,
            mean_field,
            stddev_field,
            sum_field,
            count_field,
            mode_field,
            median_field,
            vec_def_epsg,
        )

        vecDS = None
    except Exception as e:
        print("Error Vector File: {}".format(vec_file), file=sys.stderr)
        print("Error Vector Layer: {}".format(vec_lyr), file=sys.stderr)
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def calc_zonal_band_stats_test_poly_pts(
    vec_lyr_obj,
    input_img,
    img_band,
    min_thres,
    max_thres,
    out_no_data_val,
    percentile=None,
    percentile_field=None,
    min_field=None,
    max_field=None,
    mean_field=None,
    stddev_field=None,
    sum_field=None,
    count_field=None,
    mode_field=None,
    median_field=None,
    vec_def_epsg=None,
):
    """
    A function which calculates zonal statistics for a particular image band. If
    unsure then use this function. This function tests whether 1 or more pixels has
    been found within the polygon and if not then the centroid use used to find a
    value for the polygon.

    If you are unsure as to whether the pixels are small enough to be contained
    within all the polygons then use this function.

    :param vec_lyr_obj: OGR vector layer object containing the geometries being
                        processed and to which the stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band
                     will be ignored.
    :param min_thres: a lower threshold for values which will be included in
                      the stats calculation.
    :param max_thres: a upper threshold for values which will be included in
                      the stats calculation.
    :param out_no_data_val: output no data value if no valid pixels are within
                            the polygon.
    :param percentile: the percentile value to calculate (value between 0 and
                       100 inclusive).
    :param percentile_field: the name of the field for the percentile value (None or
                             not specified to be ignored).
    :param min_field: the name of the field for the min value (None or not specified
                      to be ignored).
    :param max_field: the name of the field for the max value (None or not specified
                      to be ignored).
    :param mean_field: the name of the field for the mean value (None or not specified
                       to be ignored).
    :param stddev_field: the name of the field for the standard deviation value (None
                         or not specified to be ignored).
    :param sum_field: the name of the field for the sum value (None or not specified to
                      be ignored).
    :param count_field: the name of the field for the count (of number of pixels) value
                       (None or not specified to be ignored).
    :param mode_field: the name of the field for the mode value (None or not specified
                       to be ignored).
    :param median_field: the name of the field for the median value (None or not
                         specified to be ignored).
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is the
                         projection is not well defined within the inputted
                         vector layer.

    """
    if mode_field is not None:
        import scipy.stats.mstats

    try:
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("The inputted vector layer was None")

        if (
            (min_field is None)
            and (max_field is None)
            and (mean_field is None)
            and (stddev_field is None)
            and (sum_field is None)
            and (count_field is None)
            and (mode_field is None)
            and (median_field is None)
            and (percentile_field is None)
        ):
            raise rsgislib.RSGISPyException(
                "At least one field needs to be specified for there is to an output."
            )

        imgDS = gdal.OpenEx(input_img, gdal.GA_ReadOnly)
        if imgDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(input_img))
        img_band_obj = imgDS.GetRasterBand(img_band)
        if img_band_obj is None:
            raise rsgislib.RSGISPyException(
                "Could not find image band '{}'".format(img_band)
            )
        imgGeoTrans = imgDS.GetGeoTransform()
        img_wkt_str = imgDS.GetProjection()
        img_spatial_ref = osr.SpatialReference()
        img_spatial_ref.ImportFromWkt(img_wkt_str)
        epsg_img_spatial = img_spatial_ref.GetAuthorityCode(None)

        pixel_width = imgGeoTrans[1]
        pixel_height = imgGeoTrans[5]

        imgSizeX = imgDS.RasterXSize
        imgSizeY = imgDS.RasterYSize

        imgNoDataVal = img_band_obj.GetNoDataValue()

        if vec_def_epsg is None:
            veclyr_spatial_ref = vec_lyr_obj.GetSpatialRef()
            if veclyr_spatial_ref is None:
                raise rsgislib.RSGISPyException(
                    "Could not retrieve a projection object from the vector "
                    "layer - projection might be be defined."
                )
            epsg_vec_spatial = veclyr_spatial_ref.GetAuthorityCode(None)
        else:
            epsg_vec_spatial = vec_def_epsg
            veclyr_spatial_ref = osr.SpatialReference()
            veclyr_spatial_ref.ImportFromEPSG(int(vec_def_epsg))

        if epsg_vec_spatial != epsg_img_spatial:
            imgDS = None
            vecDS = None
            raise rsgislib.RSGISPyException(
                "Inputted raster and vector layers have different "
                "projections: ('{0}' '{1}') ".format("Vector Layer Provided", input_img)
            )

        veclyrDefn = vec_lyr_obj.GetLayerDefn()

        outFieldAtts = [
            min_field,
            max_field,
            mean_field,
            stddev_field,
            sum_field,
            count_field,
            mode_field,
            median_field,
            percentile_field,
        ]
        for outattname in outFieldAtts:
            if outattname is not None:
                found = False
                for i in range(veclyrDefn.GetFieldCount()):
                    if (
                        veclyrDefn.GetFieldDefn(i).GetName().lower()
                        == outattname.lower()
                    ):
                        found = True
                        break
                if not found:
                    vec_lyr_obj.CreateField(
                        ogr.FieldDefn(outattname.lower(), ogr.OFTReal)
                    )

        fieldAttIdxs = dict()
        for outattname in outFieldAtts:
            if outattname is not None:
                fieldAttIdxs[outattname] = vec_lyr_obj.FindFieldIndex(
                    outattname.lower(), True
                )

        vec_mem_drv = ogr.GetDriverByName("Memory")
        img_mem_drv = gdal.GetDriverByName("MEM")

        # Iterate through features.
        openTransaction = False
        transactionStep = 20000
        nextTransaction = transactionStep
        nFeats = vec_lyr_obj.GetFeatureCount(True)
        pbar = tqdm.tqdm(total=nFeats)
        counter = 0
        vec_lyr_obj.ResetReading()
        feat = vec_lyr_obj.GetNextFeature()
        while feat is not None:
            if not openTransaction:
                vec_lyr_obj.StartTransaction()
                openTransaction = True

            if feat is not None:
                feat_geom = feat.geometry()
                if feat_geom is not None:
                    feat_bbox = feat_geom.GetEnvelope()

                    havepxls = True

                    x1Sp = float(feat_bbox[0] - imgGeoTrans[0])
                    x2Sp = float(feat_bbox[1] - imgGeoTrans[0])
                    y1Sp = float(feat_bbox[3] - imgGeoTrans[3])
                    y2Sp = float(feat_bbox[2] - imgGeoTrans[3])

                    if x1Sp == 0.0:
                        x1 = 0
                    else:
                        x1 = int(x1Sp / pixel_width) - 1

                    if x2Sp == 0.0:
                        x2 = 0
                    else:
                        x2 = int(x2Sp / pixel_width) + 1

                    if y1Sp == 0.0:
                        y1 = 0
                    else:
                        y1 = int(y1Sp / pixel_height) - 1

                    if y2Sp == 0.0:
                        y2 = 0
                    else:
                        y2 = int(y2Sp / pixel_height) + 1

                    if x1 < 0:
                        x1 = 0
                    elif x1 >= imgSizeX:
                        x1 = imgSizeX - 1

                    if x2 < 0:
                        x2 = 0
                    elif x2 >= imgSizeX:
                        x2 = imgSizeX - 1

                    if y1 < 0:
                        y1 = 0
                    elif y1 >= imgSizeY:
                        y1 = imgSizeY - 1

                    if y2 < 0:
                        y2 = 0
                    elif y2 >= imgSizeY:
                        y2 = imgSizeY - 1

                    xsize = x2 - x1
                    ysize = y2 - y1

                    if (xsize == 0) or (ysize == 0):
                        havepxls = False

                    # Define the image ROI for the feature
                    src_offset = (x1, y1, xsize, ysize)

                    if havepxls:
                        # Read the band array.
                        src_array = img_band_obj.ReadAsArray(*src_offset)
                    else:
                        src_array = None

                    if (src_array is not None) and havepxls:

                        # calculate new geotransform of the feature subset
                        subGeoTrans = (
                            (imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1])),
                            imgGeoTrans[1],
                            0.0,
                            (imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5])),
                            0.0,
                            imgGeoTrans[5],
                        )

                        # Create a temporary vector layer in memory
                        vec_mem_ds = vec_mem_drv.CreateDataSource("out")
                        vec_mem_lyr = vec_mem_ds.CreateLayer(
                            "poly", veclyr_spatial_ref, ogr.wkbPolygon
                        )
                        vec_mem_lyr.CreateFeature(feat.Clone())

                        # Rasterize the feature.
                        img_tmp_ds = img_mem_drv.Create(
                            "", src_offset[2], src_offset[3], 1, gdal.GDT_Byte
                        )
                        img_tmp_ds.SetGeoTransform(subGeoTrans)
                        img_tmp_ds.SetProjection(img_wkt_str)
                        gdal.RasterizeLayer(
                            img_tmp_ds, [1], vec_mem_lyr, burn_values=[1]
                        )
                        rv_array = img_tmp_ds.ReadAsArray()

                        # Mask the data vals array to feature
                        mask_arr = numpy.ones_like(src_array, dtype=numpy.uint8)
                        if imgNoDataVal is not None:
                            mask_arr[src_array == imgNoDataVal] = 0
                            mask_arr[rv_array == 0] = 0
                            mask_arr[src_array < min_thres] = 0
                            mask_arr[src_array > max_thres] = 0
                        else:
                            mask_arr[rv_array == 0] = 0
                            mask_arr[src_array < min_thres] = 0
                            mask_arr[src_array > max_thres] = 0
                        mask_arr = mask_arr.flatten()
                        src_array_flat = src_array.flatten()
                        src_array_flat = src_array_flat[mask_arr == 1]

                        if src_array_flat.shape[0] > 0:
                            if min_field is not None:
                                min_val = float(src_array_flat.min())
                                feat.SetField(fieldAttIdxs[min_field], min_val)
                            if max_field is not None:
                                max_val = float(src_array_flat.max())
                                feat.SetField(fieldAttIdxs[max_field], max_val)
                            if mean_field is not None:
                                mean_val = float(src_array_flat.mean())
                                feat.SetField(fieldAttIdxs[mean_field], mean_val)
                            if stddev_field is not None:
                                stddev_val = float(src_array_flat.std())
                                feat.SetField(fieldAttIdxs[stddev_field], stddev_val)
                            if sum_field is not None:
                                sum_val = float(src_array_flat.sum())
                                feat.SetField(fieldAttIdxs[sum_field], sum_val)
                            if count_field is not None:
                                count_val = float(src_array_flat.shape[0])
                                feat.SetField(fieldAttIdxs[count_field], count_val)
                            if mode_field is not None:
                                mode_val, mode_count = scipy.stats.mstats.mode(
                                    src_array_flat
                                )
                                mode_val = float(mode_val)
                                feat.SetField(fieldAttIdxs[mode_field], mode_val)
                            if median_field is not None:
                                median_val = float(numpy.ma.median(src_array_flat))
                                feat.SetField(fieldAttIdxs[median_field], median_val)
                            if percentile_field is not None:
                                perc_val = float(
                                    numpy.percentile(
                                        numpy.ma.compressed(src_array_flat),
                                        float(percentile),
                                    )
                                )
                                feat.SetField(fieldAttIdxs[percentile_field], perc_val)

                        else:
                            subTLX = imgGeoTrans[0] + (src_offset[0] * imgGeoTrans[1])
                            subTLY = imgGeoTrans[3] + (src_offset[1] * imgGeoTrans[5])
                            resX = imgGeoTrans[1]
                            resY = imgGeoTrans[5]

                            ptx, pty, ptz = feat.GetGeometryRef().Centroid().GetPoint()

                            xOff = math.floor((ptx - subTLX) / resX)
                            yOff = math.floor((pty - subTLY) / resY)

                            if xOff < 0:
                                xOff = 0
                            if xOff >= xsize:
                                xOff = xsize - 1

                            if yOff < 0:
                                yOff = 0
                            if yOff >= ysize:
                                yOff = ysize - 1

                            out_val = float(src_array[yOff, xOff])
                            invalid_val = False
                            if imgNoDataVal is not None:
                                if out_val == imgNoDataVal:
                                    invalid_val = True
                            if out_val < min_thres:
                                invalid_val = True
                            if out_val > max_thres:
                                invalid_val = True

                            if invalid_val:
                                if min_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[min_field], out_no_data_val
                                    )
                                if max_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[max_field], out_no_data_val
                                    )
                                if mean_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[mean_field], out_no_data_val
                                    )
                                if stddev_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[stddev_field], out_no_data_val
                                    )
                                if sum_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[sum_field], out_no_data_val
                                    )
                                if count_field is not None:
                                    feat.SetField(fieldAttIdxs[count_field], 0.0)
                                if mode_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[mode_field], out_no_data_val
                                    )
                                if median_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[median_field], out_no_data_val
                                    )
                                if percentile_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[percentile_field], out_no_data_val
                                    )
                            else:
                                if min_field is not None:
                                    feat.SetField(fieldAttIdxs[min_field], out_val)
                                if max_field is not None:
                                    feat.SetField(fieldAttIdxs[max_field], out_val)
                                if mean_field is not None:
                                    feat.SetField(fieldAttIdxs[mean_field], out_val)
                                if stddev_field is not None:
                                    feat.SetField(fieldAttIdxs[stddev_field], 0.0)
                                if sum_field is not None:
                                    feat.SetField(fieldAttIdxs[sum_field], out_val)
                                if count_field is not None:
                                    feat.SetField(fieldAttIdxs[count_field], 1.0)
                                if mode_field is not None:
                                    feat.SetField(fieldAttIdxs[mode_field], out_val)
                                if median_field is not None:
                                    feat.SetField(fieldAttIdxs[median_field], out_val)
                                if percentile_field is not None:
                                    feat.SetField(
                                        fieldAttIdxs[percentile_field], out_val
                                    )

                        # Write the updated feature to the vector layer.
                        vec_lyr_obj.SetFeature(feat)

                        vec_mem_ds = None
                        img_tmp_ds = None

            if (counter == nextTransaction) and openTransaction:
                vec_lyr_obj.CommitTransaction()
                openTransaction = False
                nextTransaction = nextTransaction + transactionStep

            feat = vec_lyr_obj.GetNextFeature()
            counter = counter + 1
            pbar.update(counter)
        if openTransaction:
            vec_lyr_obj.CommitTransaction()
            openTransaction = False
        pbar.close()

        imgDS = None
    except Exception as e:
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def ext_point_band_values_file(
    vec_file,
    vec_lyr,
    input_img,
    img_band,
    min_thres,
    max_thres,
    out_no_data_val,
    out_field,
    reproj_vec=False,
    vec_def_epsg=None,
):
    """
    A function which extracts point values for an input vector file for a
    particular image band.

    :param vec_file: input vector file
    :param vec_lyr: input vector layer within the input file which specifies the
                    features and where the output stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band
                     will be ignored.
    :param min_thres: a lower threshold for values which will be included in the
                      stats calculation.
    :param max_thres: a upper threshold for values which will be included in the
                      stats calculation.
    :param out_no_data_val: output no data value if no valid pixels are within
                            the polygon.
    :param out_field: the name of the field in the vector layer where the pixel
                      values will be written.
    :param reproj_vec: boolean to specify whether the vector layer should be
                       reprojected on the fly during processing if the projections are
                       different. Default: False to ensure it is the users intention.
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is the
                         projection is not well defined within the inputted
                         vector layer.

    """
    try:
        vecDS = gdal.OpenEx(vec_file, gdal.OF_VECTOR | gdal.OF_UPDATE)
        if vecDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

        vec_lyr_obj = vecDS.GetLayerByName(vec_lyr)
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("Could not open layer '{}'".format(vec_lyr))

        ext_point_band_values(
            vec_lyr_obj,
            input_img,
            img_band,
            min_thres,
            max_thres,
            out_no_data_val,
            out_field,
            reproj_vec,
            vec_def_epsg,
        )

        vecDS = None
    except Exception as e:
        print("Error Vector File: {}".format(vec_file), file=sys.stderr)
        print("Error Vector Layer: {}".format(vec_lyr), file=sys.stderr)
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def ext_point_band_values(
    vec_lyr_obj,
    input_img,
    img_band,
    min_thres,
    max_thres,
    out_no_data_val,
    out_field,
    reproj_vec=False,
    vec_def_epsg=None,
):
    """
    A function which extracts point values for an input vector file for a
    particular image band.

    :param vec_lyr_obj: OGR vector layer object containing the geometries being
                        processed and to which the stats will be written.
    :param input_img: the values image
    :param img_band: the index (starting at 1) of the image band for which the stats
                     will be calculated. If defined the no data value of the band
                     will be ignored.
    :param min_thres: a lower threshold for values which will be included in the
                      stats calculation.
    :param max_thres: a upper threshold for values which will be included in the
                      stats calculation.
    :param out_no_data_val: output no data value if no valid pixels are within
                            the polygon.
    :param out_field: the name of the field in the vector layer where the pixel
                      values will be written.
    :param reproj_vec: boolean to specify whether the vector layer should be
                       reprojected on the fly during processing if the projections are
                       different. Default: False to ensure it is the users intention.
    :param vec_def_epsg: an EPSG code can be specified for the vector layer is the
                         projection is not well defined within the inputted

    """
    import rsgislib.tools.geometrytools

    try:
        if vec_lyr_obj is None:
            raise rsgislib.RSGISPyException("The inputted vector layer was None")

        if out_field is None:
            raise rsgislib.RSGISPyException(
                "Output field specified as none, a name needs to be given."
            )
        elif out_field == "":
            raise rsgislib.RSGISPyException(
                "Output field specified as an empty string, a name needs to be given."
            )

        veclyrDefn = vec_lyr_obj.GetLayerDefn()
        lyr_geom_type = ogr.GeometryTypeToName(veclyrDefn.GetGeomType())
        if lyr_geom_type.lower() != "point":
            raise rsgislib.RSGISPyException("The layer geometry type must be point.")

        imgDS = gdal.OpenEx(input_img, gdal.GA_ReadOnly)
        if imgDS is None:
            raise rsgislib.RSGISPyException("Could not open '{}'".format(input_img))
        img_band_obj = imgDS.GetRasterBand(img_band)
        if img_band_obj is None:
            raise rsgislib.RSGISPyException(
                "Could not find image band '{}'".format(img_band)
            )
        imgGeoTrans = imgDS.GetGeoTransform()
        img_wkt_str = imgDS.GetProjection()
        img_spatial_ref = osr.SpatialReference()
        img_spatial_ref.ImportFromWkt(img_wkt_str)
        img_spatial_ref.AutoIdentifyEPSG()
        epsg_img_spatial = img_spatial_ref.GetAuthorityCode(None)

        pixel_width = imgGeoTrans[1]
        pixel_height = imgGeoTrans[5]

        imgSizeX = imgDS.RasterXSize
        imgSizeY = imgDS.RasterYSize

        imgNoDataVal = img_band_obj.GetNoDataValue()
        out_no_data_val = float(out_no_data_val)

        if vec_def_epsg is None:
            veclyr_spatial_ref = vec_lyr_obj.GetSpatialRef()
            if veclyr_spatial_ref is None:
                raise rsgislib.RSGISPyException(
                    "Could not retrieve a projection object from the "
                    "vector layer - projection might be be defined."
                )
            epsg_vec_spatial = veclyr_spatial_ref.GetAuthorityCode(None)
        else:
            epsg_vec_spatial = vec_def_epsg
            veclyr_spatial_ref = osr.SpatialReference()
            veclyr_spatial_ref.ImportFromEPSG(int(vec_def_epsg))
        pt_reprj = False
        if epsg_vec_spatial != epsg_img_spatial:
            if reproj_vec:
                pt_reprj = True
            else:
                raise rsgislib.RSGISPyException(
                    "Input vector and image datasets are in different "
                    "projections (EPSG:{} / EPSG:{})."
                    "You can select option to reproject.".format(
                        epsg_vec_spatial, epsg_img_spatial
                    )
                )

        found = False
        for i in range(veclyrDefn.GetFieldCount()):
            if veclyrDefn.GetFieldDefn(i).GetName().lower() == out_field.lower():
                found = True
                break
        if not found:
            vec_lyr_obj.CreateField(ogr.FieldDefn(out_field.lower(), ogr.OFTReal))

        out_field_idx = vec_lyr_obj.FindFieldIndex(out_field.lower(), True)

        # Iterate through features.
        openTransaction = False
        transactionStep = 20000
        nextTransaction = transactionStep
        nFeats = vec_lyr_obj.GetFeatureCount(True)
        pbar = tqdm.tqdm(total=nFeats)
        counter = 0
        vec_lyr_obj.ResetReading()
        feat = vec_lyr_obj.GetNextFeature()
        while feat is not None:
            if not openTransaction:
                vec_lyr_obj.StartTransaction()
                openTransaction = True

            if feat is not None:
                feat_geom = feat.geometry()
                if feat_geom is not None:
                    pt_in_img = True
                    x_pt = feat_geom.GetX()
                    y_pt = feat_geom.GetY()

                    if pt_reprj:
                        if veclyr_spatial_ref.EPSGTreatsAsLatLong():
                            x_pt, y_pt = rsgislib.tools.geometrytools.reproj_point(
                                veclyr_spatial_ref, img_spatial_ref, y_pt, x_pt
                            )
                        else:
                            x_pt, y_pt = rsgislib.tools.geometrytools.reproj_point(
                                veclyr_spatial_ref, img_spatial_ref, x_pt, y_pt
                            )

                    x_pt_off = float(x_pt - imgGeoTrans[0])
                    y_pt_off = float(y_pt - imgGeoTrans[3])

                    if x_pt_off == 0.0:
                        x_pxl = 0
                    else:
                        x_pxl = int(x_pt_off / pixel_width) - 1

                    if y_pt_off == 0.0:
                        y_pxl = 0
                    else:
                        y_pxl = int(y_pt_off / pixel_height) - 1

                    if x_pxl < 0:
                        pt_in_img = False
                    elif x_pxl >= imgSizeX:
                        pt_in_img = False

                    if y_pxl < 0:
                        pt_in_img = False
                    elif y_pxl >= imgSizeY:
                        pt_in_img = False

                    if pt_in_img:
                        src_offset = (x_pxl, y_pxl, 1, 1)
                        src_array = img_band_obj.ReadAsArray(*src_offset)
                        pxl_val = src_array[0][0]
                        out_val = float(pxl_val)
                        if pxl_val == imgNoDataVal:
                            out_val = out_no_data_val
                        elif pxl_val < min_thres:
                            out_val = out_no_data_val
                        elif pxl_val > max_thres:
                            out_val = out_no_data_val

                        feat.SetField(out_field_idx, out_val)
                    else:
                        feat.SetField(out_field_idx, out_no_data_val)

                    vec_lyr_obj.SetFeature(feat)

            if (counter == nextTransaction) and openTransaction:
                vec_lyr_obj.CommitTransaction()
                openTransaction = False
                nextTransaction = nextTransaction + transactionStep

            feat = vec_lyr_obj.GetNextFeature()
            counter = counter + 1
            pbar.update(1)
        pbar.close()

        if openTransaction:
            vec_lyr_obj.CommitTransaction()
            openTransaction = False

        imgDS = None

    except Exception as e:
        print("Error Image File: {}".format(input_img), file=sys.stderr)
        raise e


def merge_extracted_hdf5_data(h5_files, out_h5_file, datatype=None):
    """
    A function to merge a list of HDF files (e.g., from
    rsgislib.zonalstats.extractZoneImageBandValues2HDF)
    with the same number of variables (i.e., columns) into a single
    file. For example, if class training regions have been sourced
    from multiple images.

    :param h5_files: a list of input files.
    :param out_h5_file: the output file.
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    .. code:: python

        inTrainSamples = ['MSS_CloudTrain1.h5', 'MSS_CloudTrain2.h5',
                          'MSS_CloudTrain3.h5']
        cloudTrainSamples = 'LandsatMSS_CloudTrainingSamples.h5'
        rsgislib.zonalstats.merge_extracted_hdf5_data(inTrainSamples, cloudTrainSamples)

    """
    import h5py

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    first = True
    numVars = 0
    numVals = 0
    for h5File in h5_files:
        fH5 = h5py.File(h5File, "r")
        dataShp = fH5["DATA/DATA"].shape
        if first:
            numVars = dataShp[1]
            first = False
        elif numVars is not dataShp[1]:
            raise rsgislib.RSGISPyException(
                "The number of variables within the "
                "inputted HDF5 files was not the same."
            )
        numVals += dataShp[0]
        fH5.close()

    dataArr = numpy.zeros([numVals, numVars], dtype=float)

    rowInit = 0
    for h5File in h5_files:
        fH5 = h5py.File(h5File, "r")
        numRows = fH5["DATA/DATA"].shape[0]
        dataArr[rowInit : (rowInit + numRows)] = fH5["DATA/DATA"]
        rowInit += numRows
        fH5.close()

    chunk_len = 1000
    if numVals < chunk_len:
        chunk_len = numVals

    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    fH5Out = h5py.File(out_h5_file, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset(
        "DATA",
        data=dataArr,
        chunks=(chunk_len, numVars),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "Merged".encode()
    fH5Out.close()


def extract_chip_zone_image_band_values_to_hdf(
    input_image_info,
    image_mask,
    mask_value,
    chip_size,
    output_hdf,
    rotate_chips=None,
    datatype=None,
):
    """
    A function which extracts a chip/window of image pixel values. The expectation is
    that this is used to train a classifier (see deep learning functions
    in classification) but it could be used to extract image 'chips' for other purposes.

    :param input_image_info: is a list of rsgislib.imageutils.ImageBandInfo objects
                             specifying the input images and bands
    :param image_mask: is a single band input image to specify the regions of interest
    :param mask_value: is the pixel value within the imageMask to specify the
                       region of interest
    :param chip_size: is the chip size .
    :param output_hdf: is the output HDF5 file. If it all ready exists then it
                       is overwritten.
    :param rotate_chips: specify whether you wish to have the image chips rotated
                         during extraction to increase the number of samples. Default
                         is None and will therefore be ignored. Otherwise, provide
                         a list of rotation angles in degrees (e.g., [30, 60, 90,
                         120, 180])
    :param datatype: is the data type used for the output HDF5 file (e.g.,
                     rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    # Import the RIOS image reader
    from rios.imagereader import ImageReader
    import h5py
    import tqdm

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    chip_size_odd = False
    if (chip_size % 2) != 0:
        chip_size_odd = True

    chipHSize = math.floor(chip_size / 2)

    rotate = False
    n_rotations = 0
    if rotate_chips is not None:
        import scipy.ndimage

        img_win_h_size = math.floor(
            math.sqrt((chipHSize * chipHSize) + (chipHSize * chipHSize))
        )
        img_win_size = img_win_h_size * 2
        rotate = True
        n_rotations = len(rotate_chips)
        minSub = img_win_h_size - chipHSize
        maxSub = img_win_size - minSub
    else:
        img_win_h_size = chipHSize
        img_win_size = chip_size

    ######################################################################
    # Count the number of features to extract so arrays can be initialised
    # at the correct size.
    ######################################################################
    nFeats = 0
    reader = ImageReader(image_mask, windowxsize=200, windowysize=200)
    for (info, block) in tqdm.tqdm(reader):
        nFeats = nFeats + numpy.sum(block[0] == mask_value)
    ######################################################################
    if rotate:
        nFeats = nFeats * (n_rotations + 1)
        print("There are {} pixel samples (inc. rotations) in the mask.".format(nFeats))
    else:
        print("There are {} pixel samples in the mask.".format(nFeats))

    ######################################################################
    # Initialise the numpy array for the feature data
    ######################################################################
    nBands = 0
    for inImgInfo in input_image_info:
        for band in inImgInfo.bands:
            nBands = nBands + 1
    featArr = numpy.zeros([nFeats, chip_size, chip_size, nBands], dtype=numpy.float32)
    sgl_feat_arr = numpy.zeros([nBands, chip_size, chip_size], dtype=numpy.float32)
    ######################################################################

    ######################################################################
    # Populate the feature arrays with the input data
    ######################################################################
    inImgs = list()
    inImgBands = list()

    inImgs.append(image_mask)
    inImgBands.append([1])

    for inImgInfo in input_image_info:
        inImgs.append(inImgInfo.file_name)
        inImgBands.append(inImgInfo.bands)
    nImgs = len(input_image_info)

    scnOverlap = img_win_h_size

    reader = ImageReader(
        inImgs,
        windowxsize=200,
        windowysize=200,
        overlap=scnOverlap,
        layerselection=inImgBands,
    )
    iFeat = 0
    for (info, block) in tqdm.tqdm(reader):
        classMskArr = block[0]
        blkShape = classMskArr.shape

        if rotate:
            xSize = blkShape[2] - (scnOverlap * 2)
            ySize = blkShape[1] - (scnOverlap * 2)
            xRange = numpy.arange(scnOverlap, scnOverlap + xSize, 1)
            yRange = numpy.arange(scnOverlap, scnOverlap + ySize, 1)

            for y in yRange:
                yMin = y - chipHSize
                yMax = y + chipHSize
                if chip_size_odd:
                    yMax += 1
                yMinExt = y - scnOverlap
                yMaxExt = y + scnOverlap
                if chip_size_odd:
                    yMax += 1
                for x in xRange:
                    xMin = x - chipHSize
                    xMax = x + chipHSize
                    if chip_size_odd:
                        xMax += 1
                    xMinExt = x - scnOverlap
                    xMaxExt = x + scnOverlap
                    if chip_size_odd:
                        xMaxExt += 1
                    if classMskArr[0][y][x] == mask_value:
                        # Rotation 0...
                        sgl_feat_arr[...] = 0.0
                        for nImg in range(nImgs):
                            imgBlk = block[nImg + 1][..., yMin:yMax, xMin:xMax]
                            for iBand in range(imgBlk.shape[0]):
                                numpy.copyto(
                                    sgl_feat_arr[iBand], imgBlk[iBand], casting="safe"
                                )
                        numpy.copyto(featArr[iFeat], sgl_feat_arr.T, casting="safe")
                        iFeat = iFeat + 1
                        # Iterate through rotation angles
                        for rotate_angle in rotate_chips:
                            # Perform Rotate
                            sgl_feat_arr[...] = 0.0
                            for nImg in range(nImgs):
                                imgBlk = block[nImg + 1][
                                    ..., yMinExt:yMaxExt, xMinExt:xMaxExt
                                ]
                                # Perform Rotate
                                imgBlkRot = scipy.ndimage.rotate(
                                    imgBlk,
                                    rotate_angle,
                                    axes=[1, 2],
                                    reshape=False,
                                    output=numpy.float32,
                                    mode="nearest",
                                )
                                for iBand in range(imgBlk.shape[0]):
                                    numpy.copyto(
                                        sgl_feat_arr[iBand],
                                        imgBlkRot[iBand, minSub:maxSub, minSub:maxSub],
                                        casting="safe",
                                    )
                            numpy.copyto(featArr[iFeat], sgl_feat_arr.T, casting="safe")
                            iFeat = iFeat + 1
        else:
            xSize = blkShape[2] - (scnOverlap * 2)
            ySize = blkShape[1] - (scnOverlap * 2)
            xRange = numpy.arange(scnOverlap, scnOverlap + xSize, 1)
            yRange = numpy.arange(scnOverlap, scnOverlap + ySize, 1)

            for y in yRange:
                yMin = y - scnOverlap
                yMax = y + scnOverlap
                if chip_size_odd:
                    yMax += 1
                for x in xRange:
                    xMin = x - scnOverlap
                    xMax = x + scnOverlap
                    if chip_size_odd:
                        xMax += 1
                    if classMskArr[0][y][x] == mask_value:
                        sgl_feat_arr[...] = 0.0
                        for nImg in range(nImgs):
                            imgBlk = block[nImg + 1][..., yMin:yMax, xMin:xMax]
                            for iBand in range(imgBlk.shape[0]):
                                numpy.copyto(
                                    sgl_feat_arr[iBand], imgBlk[iBand], casting="safe"
                                )
                        numpy.copyto(featArr[iFeat], sgl_feat_arr.T, casting="safe")
                        iFeat = iFeat + 1
    ######################################################################

    ######################################################################
    # Create the output HDF5 file and populate with data.
    ######################################################################
    fH5Out = h5py.File(output_hdf, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    # Chunk size needs to be less than number of data points
    if nFeats < 250:
        chunkFeatures = nFeats
    else:
        chunkFeatures = 250
    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)
    dataGrp.create_dataset(
        "DATA",
        data=featArr,
        chunks=(chunkFeatures, chip_size, chip_size, nBands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "IMAGE TILES".encode()
    fH5Out.close()
    ######################################################################


def split_sample_chip_hdf5_file(
    input_h5_file, sample_h5_file, remain_h5_file, sample_size, rnd_seed, datatype=None
):
    """
    A function to split the HDF5 outputs from the
    rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf
    function into two sets by taking a random set with the defined sample
    size from the input file, saving the sample and the remainder to output HDF5 files.

    :param input_h5_file: The input HDF5 file to the split.
    :param sample_h5_file: The output HDF5 file with the sample outputted.
    :param remain_h5_file: The output HDF5 file with the remainder outputted.
    :param sample_size: An integer specifying the size of the sample to be taken.
    :param rnd_seed: An integer specifying the seed for the random number generator,
                     allowing the same 'random' sample to be taken.
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import h5py

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    f = h5py.File(input_h5_file, "r")
    n_rows = f["DATA/DATA"].shape[0]
    chip_size = f["DATA/DATA"].shape[1]
    n_bands = f["DATA/DATA"].shape[3]
    f.close()

    if sample_size > n_rows:
        raise rsgislib.RSGISPyException(
            "The requested sample is larger than the number samples in the input file."
        )

    rnd_obj = numpy.random.RandomState(rnd_seed)
    # Find sufficient unique sample indexes.
    smp_idxs = numpy.zeros(sample_size, dtype=int)
    while numpy.unique(smp_idxs).shape[0] != sample_size:
        tmp_idxs = rnd_obj.randint(0, n_rows, sample_size)
        tmp_uniq_idxs = numpy.unique(tmp_idxs)
        c_idx = 0
        if numpy.sum(smp_idxs) > 0:
            c_smp = smp_idxs[smp_idxs > 0]
            tmp_idxs = numpy.concatenate((c_smp, tmp_uniq_idxs))
            tmp_uniq_idxs = numpy.unique(tmp_idxs)

        max_idx = sample_size - 1
        if tmp_uniq_idxs.shape[0] < max_idx:
            max_idx = tmp_uniq_idxs.shape[0]
        smp_idxs[0:max_idx] = tmp_uniq_idxs[0:max_idx]
    smp_idxs = numpy.sort(smp_idxs)

    # Get the remaining indexes
    remain_idxs = numpy.arange(0, n_rows)
    remain_idxs = remain_idxs[
        numpy.isin(remain_idxs, smp_idxs, assume_unique=True, invert=True)
    ]

    # Read the input HDF5 file.
    f = h5py.File(input_h5_file, "r")
    in_samples = f["DATA/DATA"]
    out_samples = in_samples[smp_idxs]
    remain_samples = in_samples[remain_idxs]
    f.close()

    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    # Create an output HDF5 file and populate with sample data.
    if sample_size < 250:
        sample_chunks = sample_size
    else:
        sample_chunks = 250
    fSampleH5Out = h5py.File(sample_h5_file, "w")
    dataSampleGrp = fSampleH5Out.create_group("DATA")
    metaSampleGrp = fSampleH5Out.create_group("META-DATA")
    dataSampleGrp.create_dataset(
        "DATA",
        data=out_samples,
        chunks=(sample_chunks, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describSampleDS = metaSampleGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describSampleDS[0] = "IMAGE TILES".encode()
    fSampleH5Out.close()

    # Create an output HDF5 file and populate with remain data.
    if (n_rows - sample_size) < 250:
        sample_chunks = n_rows - sample_size
    else:
        sample_chunks = 250
    fSampleH5Out = h5py.File(remain_h5_file, "w")
    dataSampleGrp = fSampleH5Out.create_group("DATA")
    metaSampleGrp = fSampleH5Out.create_group("META-DATA")
    dataSampleGrp.create_dataset(
        "DATA",
        data=remain_samples,
        chunks=(sample_chunks, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describSampleDS = metaSampleGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describSampleDS[0] = "IMAGE TILES".encode()
    fSampleH5Out.close()


def merge_extracted_hdf5_chip_data(h5_files, out_h5_file, datatype=None):
    """
    A function to merge a list of HDF files
    (e.g., from rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf)
    with the same number of variables (i.e., image bands) and chip size into
    a single file. For example, if class training regions have been sourced
    from multiple images.

    :param h5_files: a list of input files.
    :param out_h5_file: the output file.
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    .. code:: python

        inTrainSamples = ['MSS_CloudTrain1.h5', 'MSS_CloudTrain2.h5',
                          'MSS_CloudTrain3.h5']
        cloudTrainSamples = 'LandsatMSS_CloudTrainingSamples.h5'
        rsgislib.zonalstats.merge_extracted_hdf5_chip_data(inTrainSamples,
                                                           cloudTrainSamples)

    """
    import h5py

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    first = True
    n_feats = 0
    chip_size = 0
    n_bands = 0
    for h5File in h5_files:
        fH5 = h5py.File(h5File, "r")
        data_shp = fH5["DATA/DATA"].shape
        if first:
            n_bands = data_shp[3]
            chip_size = data_shp[1]
            first = False
        else:
            if n_bands != data_shp[3]:
                raise rsgislib.RSGISPyException(
                    "The number of bands (variables) within the inputted "
                    "HDF5 files was not the same."
                )
            if chip_size != data_shp[1]:
                raise rsgislib.RSGISPyException(
                    "The chip size within the inputted HDF5 files was not the same."
                )
        n_feats += data_shp[0]
        fH5.close()

    feat_arr = numpy.zeros(
        [n_feats, chip_size, chip_size, n_bands], dtype=numpy.float32
    )

    row_init = 0
    for h5File in h5_files:
        fH5 = h5py.File(h5File, "r")
        n_rows = fH5["DATA/DATA"].shape[0]
        feat_arr[row_init : (row_init + n_rows)] = fH5["DATA/DATA"]
        row_init += n_rows
        fH5.close()

    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    chunk_size = 250
    if n_feats < chunk_size:
        chunk_size = n_feats

    fH5Out = h5py.File(out_h5_file, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset(
        "DATA",
        data=feat_arr,
        chunks=(chunk_size, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "Merged".encode()
    fH5Out.close()


def extract_ref_chip_zone_image_band_values_to_hdf(
    input_image_info,
    ref_img,
    ref_img_band,
    image_mask,
    mask_value,
    chip_size,
    output_hdf,
    rotate_chips=None,
    datatype=None,
):
    """
    A function which extracts a chip/window of image pixel values. The expectation is
    that this is used to train a classifier (see deep learning functions in
    classification) but it could be used to extract image 'chips' for other purposes.

    :param input_image_info: is a list of rsgislib.imageutils.ImageBandInfo objects
                             specifying the input images and bands
    :param ref_img: is an image file (same pixel size and projection as the other
                    input images) which is used as the class training
    :param ref_img_band: is the image band in the reference image to be used (only a
                         single reference band can be used).
    :param image_mask: is a single band input image to specify the regions of interest
    :param mask_value: is the pixel value within the imageMask to specify the
                       region of interest
    :param chip_size: is the chip size .
    :param output_hdf: is the output HDF5 file. If it all ready exists then it
                       is overwritten.
    :param rotate_chips: specify whether you wish to have the image chips rotated
                         during extraction to increase the number of samples. Default
                         is None and will therefore be ignored. Otherwise, provide a
                         list of rotation angles in degrees (e.g., [30, 60, 90,
                         120, 180])
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    # Import the RIOS image reader
    from rios.imagereader import ImageReader
    import h5py
    import tqdm

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    if (chip_size % 2) != 0:
        raise rsgislib.RSGISPyException("The chip size must be an even number.")

    chipHSize = math.floor(chip_size / 2)

    rotate = False
    n_rotations = 0
    if rotate_chips is not None:
        import scipy.ndimage

        img_win_h_size = math.floor(
            math.sqrt((chipHSize * chipHSize) + (chipHSize * chipHSize))
        )
        img_win_size = img_win_h_size * 2
        rotate = True
        n_rotations = len(rotate_chips)
        minSub = img_win_h_size - chipHSize
        maxSub = img_win_size - minSub
    else:
        img_win_h_size = chipHSize
        img_win_size = chip_size

    ######################################################################
    # Count the number of features to extract so arrays can be initialised
    # at the correct size.
    ######################################################################
    nFeats = 0
    reader = ImageReader(image_mask, windowxsize=200, windowysize=200)
    for (info, block) in tqdm.tqdm(reader):
        nFeats = nFeats + numpy.sum(block[0] == mask_value)
    ######################################################################
    if rotate:
        nFeats = nFeats * (n_rotations + 1)
        print("There are {} pixel samples (inc. rotations) in the mask.".format(nFeats))
    else:
        print("There are {} pixel samples in the mask.".format(nFeats))

    ######################################################################
    # Initialise the numpy array for the feature data
    ######################################################################
    nBands = 0
    for inImgInfo in input_image_info:
        for band in inImgInfo.bands:
            nBands = nBands + 1
    featArr = numpy.zeros(
        [nFeats, chip_size, chip_size, nBands], dtype=numpy.float32
    )  # [nFeats, chipSize, chipSize, nBands]
    sgl_feat_arr = numpy.zeros(
        [nBands, chip_size, chip_size], dtype=numpy.float32
    )  # [chipSize, chipSize, nBands]
    featRefArr = numpy.zeros([nFeats, chip_size, chip_size], dtype=numpy.uint16)
    ######################################################################

    ######################################################################
    # Populate the feature arrays with the input data
    ######################################################################
    inImgs = list()
    inImgBands = list()

    inImgs.append(image_mask)
    inImgBands.append([1])
    inImgs.append(ref_img)
    inImgBands.append([ref_img_band])

    for inImgInfo in input_image_info:
        inImgs.append(inImgInfo.file_name)
        inImgBands.append(inImgInfo.bands)
    nImgs = len(input_image_info)

    scnOverlap = img_win_h_size

    reader = ImageReader(
        inImgs,
        windowxsize=200,
        windowysize=200,
        overlap=scnOverlap,
        layerselection=inImgBands,
    )
    iFeat = 0
    for (info, block) in tqdm.tqdm(reader):
        classMskArr = block[0]
        blkShape = classMskArr.shape

        if rotate:
            xSize = blkShape[2] - (scnOverlap * 2)
            ySize = blkShape[1] - (scnOverlap * 2)
            xRange = numpy.arange(scnOverlap, scnOverlap + xSize, 1)
            yRange = numpy.arange(scnOverlap, scnOverlap + ySize, 1)

            for y in yRange:
                yMin = y - chipHSize
                yMax = y + chipHSize
                yMinExt = y - scnOverlap
                yMaxExt = y + scnOverlap
                for x in xRange:
                    xMin = x - chipHSize
                    xMax = x + chipHSize
                    xMinExt = x - scnOverlap
                    xMaxExt = x + scnOverlap
                    if classMskArr[0][y][x] == mask_value:
                        # Rotation 0...
                        refImgBlk = block[1][0, yMin:yMax, xMin:xMax]
                        numpy.copyto(featRefArr[iFeat], refImgBlk, casting="safe")
                        sgl_feat_arr[...] = 0.0
                        for nImg in range(nImgs):
                            imgBlk = block[nImg + 2][..., yMin:yMax, xMin:xMax]
                            for iBand in range(imgBlk.shape[0]):
                                numpy.copyto(
                                    sgl_feat_arr[iBand], imgBlk[iBand], casting="safe"
                                )
                        numpy.copyto(featArr[iFeat], sgl_feat_arr.T, casting="safe")
                        iFeat = iFeat + 1
                        # Iterate through rotation angles
                        for rotate_angle in rotate_chips:
                            refImgBlk = block[1][0, yMinExt:yMaxExt, xMinExt:xMaxExt]
                            # Perform Rotate
                            refImgBlkRot = scipy.ndimage.rotate(
                                refImgBlk,
                                rotate_angle,
                                reshape=False,
                                output=numpy.uint16,
                                order=0,
                                mode="nearest",
                            )
                            numpy.copyto(
                                featRefArr[iFeat],
                                refImgBlkRot[minSub:maxSub, minSub:maxSub],
                                casting="safe",
                            )
                            sgl_feat_arr[...] = 0.0
                            for nImg in range(nImgs):
                                imgBlk = block[nImg + 2][
                                    ..., yMinExt:yMaxExt, xMinExt:xMaxExt
                                ]
                                # Perform Rotate
                                imgBlkRot = scipy.ndimage.rotate(
                                    imgBlk,
                                    rotate_angle,
                                    axes=[1, 2],
                                    reshape=False,
                                    output=numpy.float32,
                                    mode="nearest",
                                )
                                for iBand in range(imgBlk.shape[0]):
                                    numpy.copyto(
                                        sgl_feat_arr[iBand],
                                        imgBlkRot[iBand, minSub:maxSub, minSub:maxSub],
                                        casting="safe",
                                    )
                            numpy.copyto(featArr[iFeat], sgl_feat_arr.T, casting="safe")
                            iFeat = iFeat + 1
        else:
            xSize = blkShape[2] - (scnOverlap * 2)
            ySize = blkShape[1] - (scnOverlap * 2)
            xRange = numpy.arange(scnOverlap, scnOverlap + xSize, 1)
            yRange = numpy.arange(scnOverlap, scnOverlap + ySize, 1)

            for y in yRange:
                yMin = y - scnOverlap
                yMax = y + scnOverlap
                for x in xRange:
                    xMin = x - scnOverlap
                    xMax = x + scnOverlap
                    if classMskArr[0][y][x] == mask_value:
                        refImgBlk = block[1][0, yMin:yMax, xMin:xMax]
                        numpy.copyto(featRefArr[iFeat], refImgBlk, casting="safe")
                        sgl_feat_arr[...] = 0.0
                        for nImg in range(nImgs):
                            imgBlk = block[nImg + 2][..., yMin:yMax, xMin:xMax]
                            for iBand in range(imgBlk.shape[0]):
                                numpy.copyto(
                                    sgl_feat_arr[iBand], imgBlk[iBand], casting="safe"
                                )
                        numpy.copyto(featArr[iFeat], sgl_feat_arr.T, casting="safe")
                        iFeat = iFeat + 1
    ######################################################################

    ######################################################################
    # Create the output HDF5 file and populate with data.
    ######################################################################
    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    fH5Out = h5py.File(output_hdf, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    # Chunk size needs to be less than number of data points
    if nFeats < 250:
        chunkFeatures = nFeats
    else:
        chunkFeatures = 250
    dataGrp.create_dataset(
        "DATA",
        data=featArr,
        chunks=(chunkFeatures, chip_size, chip_size, nBands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    dataGrp.create_dataset(
        "REF",
        data=featRefArr,
        chunks=(chunkFeatures, chip_size, chip_size),
        compression="gzip",
        shuffle=True,
        dtype="H",
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "IMAGE REF TILES".encode()
    fH5Out.close()
    ######################################################################


def split_sample_ref_chip_hdf5_file(
    input_h5_file, sample_h5_file, remain_h5_file, sample_size, rnd_seed, datatype=None
):
    """
    A function to split the HDF5 outputs from the
    rsgislib.zonalstats.extract_chip_zone_image_band_values_to_hdf
    function into two sets by taking a random set with the defined sample size
    from the input file, saving the sample and the remainder to output HDF5 files.

    :param input_h5_file: The input HDF5 file to the split.
    :param sample_h5_file: The output HDF5 file with the sample outputted.
    :param remain_h5_file: The output HDF5 file with the remainder outputted.
    :param sample_size: An integer specifying the size of the sample to be taken.
    :param rnd_seed: An integer specifying the seed for the random number generator,
                     allowing the same 'random' sample to be taken.
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    """
    import h5py

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    f = h5py.File(input_h5_file, "r")
    n_rows = f["DATA/REF"].shape[0]
    chip_size = f["DATA/REF"].shape[1]
    n_bands = f["DATA/DATA"].shape[3]
    f.close()

    if sample_size > n_rows:
        raise rsgislib.RSGISPyException(
            "The requested sample is larger than the number samples in the input file."
        )

    rnd_obj = numpy.random.RandomState(rnd_seed)
    # Find sufficient unique sample indexes.
    smp_idxs = numpy.zeros(sample_size, dtype=int)
    while numpy.unique(smp_idxs).shape[0] != sample_size:
        tmp_idxs = rnd_obj.randint(0, n_rows, sample_size)
        tmp_uniq_idxs = numpy.unique(tmp_idxs)
        c_idx = 0
        if numpy.sum(smp_idxs) > 0:
            c_smp = smp_idxs[smp_idxs > 0]
            tmp_idxs = numpy.concatenate((c_smp, tmp_uniq_idxs))
            tmp_uniq_idxs = numpy.unique(tmp_idxs)

        max_idx = sample_size - 1
        if tmp_uniq_idxs.shape[0] < max_idx:
            max_idx = tmp_uniq_idxs.shape[0]
        smp_idxs[0:max_idx] = tmp_uniq_idxs[0:max_idx]
    smp_idxs = numpy.sort(smp_idxs)

    # Get the remaining indexes
    remain_idxs = numpy.arange(0, n_rows)
    remain_idxs = remain_idxs[
        numpy.isin(remain_idxs, smp_idxs, assume_unique=True, invert=True)
    ]

    # Read the input HDF5 file.
    f = h5py.File(input_h5_file, "r")
    in_data_samples = f["DATA/DATA"]
    out_data_samples = in_data_samples[smp_idxs]
    remain_data_samples = in_data_samples[remain_idxs]
    in_ref_samples = f["DATA/REF"]
    out_ref_samples = in_ref_samples[smp_idxs]
    remain_ref_samples = in_ref_samples[remain_idxs]
    reman_smpl_size = remain_ref_samples.shape[0]
    f.close()

    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    # Create an output HDF5 file and populate with sample data.
    fSampleH5Out = h5py.File(sample_h5_file, "w")
    dataSampleGrp = fSampleH5Out.create_group("DATA")
    metaSampleGrp = fSampleH5Out.create_group("META-DATA")
    smpl_chk_size = 250
    if sample_size < smpl_chk_size:
        smpl_chk_size = sample_size
    dataSampleGrp.create_dataset(
        "DATA",
        data=out_data_samples,
        chunks=(smpl_chk_size, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    dataSampleGrp.create_dataset(
        "REF",
        data=out_ref_samples,
        chunks=(smpl_chk_size, chip_size, chip_size),
        compression="gzip",
        shuffle=True,
        dtype="H",
    )
    describSampleDS = metaSampleGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describSampleDS[0] = "IMAGE REF TILES".encode()
    fSampleH5Out.close()

    # Create an output HDF5 file and populate with remain data.
    fSampleH5Out = h5py.File(remain_h5_file, "w")
    dataSampleGrp = fSampleH5Out.create_group("DATA")
    metaSampleGrp = fSampleH5Out.create_group("META-DATA")
    rman_chk_size = 250
    if reman_smpl_size < rman_chk_size:
        rman_chk_size = reman_smpl_size
    dataSampleGrp.create_dataset(
        "DATA",
        data=remain_data_samples,
        chunks=(rman_chk_size, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    dataSampleGrp.create_dataset(
        "REF",
        data=remain_ref_samples,
        chunks=(rman_chk_size, chip_size, chip_size),
        compression="gzip",
        shuffle=True,
        dtype="H",
    )
    describSampleDS = metaSampleGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describSampleDS[0] = "IMAGE REF TILES".encode()
    fSampleH5Out.close()


def merge_extracted_hdf5_chip_ref_data(h5_files, out_h5_file, datatype=None):
    """
    A function to merge a list of HDF files
    (e.g., from rsgislib.zonalstats.extract_ref_chip_zone_image_band_values_to_hdf)
    with the same number of variables (i.e., image bands) and chip size into
    a single file. For example, if class training regions have been sourced
    from multiple images.

    :param h5_files: a list of input files.
    :param out_h5_file: the output file.
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.

    .. code:: python

        inTrainSamples = ['MSS_CloudTrain1.h5', 'MSS_CloudTrain2.h5',
                          'MSS_CloudTrain3.h5']
        cloudTrainSamples = 'LandsatMSS_CloudTrainingSamples.h5'
        rsgislib.zonalstats.merge_extracted_hdf5_chip_ref_data(inTrainSamples,
                                                               cloudTrainSamples)

    """
    import h5py

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT

    first = True
    n_feats = 0
    chip_size = 0
    n_bands = 0
    for h5File in h5_files:
        fH5 = h5py.File(h5File, "r")
        data_shp = fH5["DATA/DATA"].shape
        if first:
            n_bands = data_shp[3]
            chip_size = data_shp[1]
            first = False
        else:
            if n_bands != data_shp[3]:
                raise rsgislib.RSGISPyException(
                    "The number of bands (variables) within the inputted "
                    "HDF5 files was not the same."
                )
            if chip_size != data_shp[1]:
                raise rsgislib.RSGISPyException(
                    "The chip size within the inputted HDF5 files was not the same."
                )
        n_feats += data_shp[0]
        fH5.close()

    feat_arr = numpy.zeros(
        [n_feats, chip_size, chip_size, n_bands], dtype=numpy.float32
    )
    feat_ref_arr = numpy.zeros([n_feats, chip_size, chip_size], dtype=numpy.uint16)

    row_init = 0
    for h5File in h5_files:
        fH5 = h5py.File(h5File, "r")
        n_rows = fH5["DATA/DATA"].shape[0]
        feat_arr[row_init : (row_init + n_rows)] = fH5["DATA/DATA"]
        feat_ref_arr[row_init : (row_init + n_rows)] = fH5["DATA/REF"]
        row_init += n_rows
        fH5.close()

    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    chunk_size = 250
    if n_feats < chunk_size:
        chunk_size = n_feats

    fH5Out = h5py.File(out_h5_file, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset(
        "DATA",
        data=feat_arr,
        chunks=(chunk_size, chip_size, chip_size, n_bands),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    dataGrp.create_dataset(
        "REF",
        data=feat_ref_arr,
        chunks=(chunk_size, chip_size, chip_size),
        compression="gzip",
        shuffle=True,
        dtype="H",
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "Merged".encode()
    fH5Out.close()


def msk_h5_smpls_to_finite_values(
    input_h5, output_h5, datatype=None, lower_limit=None, upper_limit=None
):
    """
    A function to remove values from a HDF5 sample file which are not finite.
    Upper and lower values can also be specified.

    :param input_h5: Input HDF5 file.
    :param output_h5: Output HDF5 file.
    :param datatype: is the data type used for the output HDF5 file
                     (e.g., rsgislib.TYPE_32FLOAT). If None (default)
                     then the output data type will be float32.
    :param lower_limit: Optional lower value threshold (if None then not used).
    :param upper_limit: Optional upper value threshold (if None then not used).

    """
    import h5py

    if datatype is None:
        datatype = rsgislib.TYPE_32FLOAT
    h5_dtype = rsgislib.get_numpy_char_codes_datatype(datatype)

    fH5 = h5py.File(input_h5, "r")
    data_shp = fH5["DATA/DATA"].shape
    num_vars = data_shp[1]
    data = numpy.array(fH5["DATA/DATA"])
    data = data[numpy.isfinite(data).all(axis=1)]
    if lower_limit is not None:
        data = data[numpy.any(data > lower_limit, axis=1)]
    if upper_limit is not None:
        data = data[numpy.any(data < upper_limit, axis=1)]

    n_samples = data.shape[0]
    chunk_size = 1000
    if n_samples < 1000:
        chunk_size = n_samples

    fH5Out = h5py.File(output_h5, "w")
    dataGrp = fH5Out.create_group("DATA")
    metaGrp = fH5Out.create_group("META-DATA")
    dataGrp.create_dataset(
        "DATA",
        data=data,
        chunks=(chunk_size, num_vars),
        compression="gzip",
        shuffle=True,
        dtype=h5_dtype,
    )
    describDS = metaGrp.create_dataset("DESCRIPTION", (1,), dtype="S10")
    describDS[0] = "finite values".encode()
    fH5Out.close()


def get_var_from_hdf5_data(h5_files: List[str], var_idx: int = 0) -> numpy.array:
    """
    A function to get the data for a specific variable from a list of HDF files
     (e.g., from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf)

    :param h5_files: a list of input files.
    :param var_idx: the index for the variable of interest. Note array indexing
                    starts at 0. So if you want image band 2 then that will be
                    index 1 etc.
    :return: numpy array with the data or None is there is no data to return.

    """
    import h5py

    if var_idx < 0:
        raise rsgislib.RSGISPyException("The variable index must be greater than 0.")

    num_vals = 0
    for h5_file in h5_files:
        f_obj_h5 = h5py.File(h5_file, "r")
        data_shp = f_obj_h5["DATA/DATA"].shape
        if var_idx < data_shp[1]:
            num_vals += data_shp[0]
        f_obj_h5.close()

    if num_vals == 0:
        return None

    data_arr = numpy.zeros(num_vals, dtype=float)

    row_init = 0
    for h5_file in h5_files:
        f_obj_h5 = h5py.File(h5_file, "r")
        data_shp = f_obj_h5["DATA/DATA"].shape
        if var_idx < data_shp[1]:
            num_rows = f_obj_h5["DATA/DATA"].shape[0]
            data_arr[row_init : (row_init + num_rows)] = f_obj_h5["DATA/DATA"][
                ..., var_idx
            ]
            row_init += num_rows
        f_obj_h5.close()

    return data_arr


def get_hdf5_data(h5_files: List[str]) -> numpy.array:
    """
    A function to get the data from a list of HDF files
     (e.g., from rsgislib.zonalstats.extract_zone_img_band_values_to_hdf)

    :param h5_files: a list of input files.
    :return: numpy array with the data or None is there is no data to return.

    """
    import h5py

    num_vals = 0
    n_vars = 0
    first = True
    for h5_file in h5_files:
        f_obj_h5 = h5py.File(h5_file, "r")
        data_shp = f_obj_h5["DATA/DATA"].shape
        if first:
            n_vars = data_shp[1]
            first = False
        else:
            if n_vars != data_shp[1]:
                raise rsgislib.RSGISPyException(
                    "The input hdf5 files have a different number of input variables"
                )
        num_vals += data_shp[0]
        f_obj_h5.close()

    if num_vals == 0:
        return None

    data_arr = numpy.zeros([num_vals, n_vars], dtype=float)

    row_init = 0
    for h5_file in h5_files:
        f_obj_h5 = h5py.File(h5_file, "r")
        num_rows = f_obj_h5["DATA/DATA"].shape[0]
        data_arr[row_init : (row_init + num_rows)] = f_obj_h5["DATA/DATA"]
        row_init += num_rows
        f_obj_h5.close()

    return data_arr



#! /usr/bin/env python

############################################################################
# Copyright (c) 2016 Dr. Peter Bunting, Aberystwyth University
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#
# Purpose:  A command to convert an input image to a KMZ file. 
#
# Author: Pete Bunting
# Email: pfb@aber.ac.uk
# Date: 14/04/16
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
#############################################################################


import rsgislib.tools.visualisation
import argparse

parser = argparse.ArgumentParser(prog='rsgisimg2kmz.py', description='''Create a KMZ file of the input image...''')

parser.add_argument("-i", "--input", type=str, required=True, help='Input image.')
parser.add_argument("-o", "--output", type=str, required=True, help='Output KMZ.')
parser.add_argument("-b", "--bands", type=str, required=True, help='Comma seperated set of bands. Either single or r,g,b.')
parser.add_argument("-n", "--noproj", action='store_false', default=True, help='Do not re-project the image to lat/long.')
parser.add_argument("-f", "--finite", action='store_true', default=False, help='Apply a finite mask to the input image before stretching.')

# Call the parser to parse the arguments.
args = parser.parse_args()

rsgislib.tools.visualisation.create_kmz_img(args.input, args.output, args.bands, args.noproj, args.finite)




import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
SEGMENT_DATA_DIR = os.path.join(DATA_DIR, "segment")
IMGCALC_DATA_DIR = os.path.join(DATA_DIR, "imagecalc")


def test_clump(tmp_path):
    import rsgislib.segmentation.tiledclump

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    clumps_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.segmentation.clump(
        input_img,
        clumps_img,
        gdalformat="KEA",
        in_memory=False,
        no_data_val=0,
        add_to_rat=False,
    )
    assert os.path.exists(clumps_img)


# TODO rsgislib.segmentation.label_pixels_from_cluster_centres
# TODO rsgislib.segmentation.relabel_clumps
# TODO rsgislib.segmentation.eliminate_single_pixels
# TODO rsgislib.segmentation.rm_small_clumps
# TODO rsgislib.segmentation.rm_small_clumps_stepwise


def test_union_of_clumps(tmp_path):
    import rsgislib.segmentation.tiledclump

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    clumps_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.segmentation.union_of_clumps(
        [input_img, input_img],
        clumps_img,
        gdalformat="KEA",
        no_data_val=0,
        add_to_rat=False,
    )
    assert os.path.exists(clumps_img)


# TODO rsgislib.segmentation.mean_image
# TODO rsgislib.segmentation.merge_segmentation_tiles


def test_generate_regular_grid(tmp_path):
    import rsgislib.segmentation.tiledclump

    input_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi_cats.kea")
    clumps_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.segmentation.generate_regular_grid(
        input_img, clumps_img, "KEA", 500, 500, 250
    )
    assert os.path.exists(clumps_img)


# TODO rsgislib.segmentation.drop_selected_clumps
# TODO rsgislib.segmentation.find_tile_borders_mask
# TODO rsgislib.segmentation.include_regions_in_clumps
# TODO rsgislib.segmentation.merge_clump_images
# TODO rsgislib.segmentation.merge_equiv_clumps
# TODO rsgislib.segmentation.merge_segments_to_neighbours



#!/usr/bin/env python

#############################################
# rsgisimginfo.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
# Purpose:  Get information on vector files/layers. 
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 02/01/2019
# Version: 1.0
# 
#############################################

import argparse
import rsgislib.rastergis
import rsgislib.imageutils

if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Print image info to console.")
    parser.add_argument("-i", "--img", type=str, required=True, help="image file")
    args = parser.parse_args()

    print(f"Image: {args.img}")
    img_size = rsgislib.imageutils.get_img_size(args.img)
    print(f"Size: {img_size[0]} x {img_size[1]}")
    img_pxl_res = rsgislib.imageutils.get_img_res(args.img)
    print(f"Pixel Size: {img_pxl_res[0]} x {img_pxl_res[1]}")
    img_epsg = rsgislib.imageutils.get_epsg_proj_from_img(args.img)
    print(f"Image EPSG: {img_epsg}")
    n_bands = rsgislib.imageutils.get_img_band_count(args.img)
    print(f"N Bands: {n_bands}")

    band_names = rsgislib.imageutils.get_band_names(args.img)
    for band, band_name in enumerate(band_names):
        print(f"\tBand {band+1}: {band_name}")
        if rsgislib.imageutils.is_img_thematic(args.img, band+1):
            rat_cols = rsgislib.rastergis.get_rat_columns(args.img, band+1)
            print(f"\t RAT cols: {rat_cols}")


